from fpdf import FPDF
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
import numpy as np
from PIL import Image

fig = Figure(figsize=(6, 4), dpi=300)
fig.subplots_adjust(top=0.8)
ax1 = fig.add_subplot(211)
ax1.set_ylabel("volts")
ax1.set_title("a sine wave")

t = np.arange(0.0, 1.0, 0.01)
s = np.sin(2 * np.pi * t)
(line,) = ax1.plot(t, s, color="blue", lw=2)

# Fixing random state for reproducibility
np.random.seed(19680801)

ax2 = fig.add_axes([0.15, 0.1, 0.7, 0.3])
n, bins, patches = ax2.hist(
    np.random.randn(1000), 50, facecolor="yellow", edgecolor="yellow"
)
ax2.set_xlabel("time (s)")

# Converting Figure to an image:
canvas = FigureCanvas(fig)
canvas.draw()
img = Image.fromarray(np.asarray(canvas.buffer_rgba()))

pdf = FPDF()
pdf.add_page()
pdf.image(img, w=pdf.epw)  # Make the image full width
pdf.output("matplotlib.pdf")



from pathlib import Path

import pytest

from fpdf import FPDF, FPDFException
from fpdf.enums import MethodReturnValue, YPos

from test.conftest import assert_pdf_equal, LOREM_IPSUM, assert_same_file

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"

TEXT_SIZE, SPACING = 36, 1.15
LINE_HEIGHT = TEXT_SIZE * SPACING

TABLE_DATA = (
    ("First name", "Last name", "Age", "City"),
    ("Jules", "Smith", "34", "San Juan"),
    ("Mary", "Ramos", "45", "Orlando"),
    ("Carlson", "Banks", "19", "Los Angeles"),
    ("Lucas", "Cimon", "31", "Angers"),
)

LONG_TEXT = (
    "\nProfessor: (Eric Idle) It's an entirely new strain of sheep, a killer sheep that can not only hold a rifle but is also a first-class shot.\n"
    "Assistant: But where are they coming from, professor?\n"
    "Professor: That I don't know. I just don't know. I really just don't know. I'm afraid I really just don't know. I'm afraid even I really just"
    " don't know. I have to tell you I'm afraid even I really just don't know. I'm afraid I have to tell you... (she hands him a glass of water"
    " which she had been busy getting as soon as he started into this speech) ... thank you ... (resuming normal breezy voice) ... I don't know."
    " Our only clue is this portion of wolf's clothing which the killer sheep ..."
)


def test_multi_cell_page_break_with_fill(tmp_path):
    # Tests condition from issue #1471
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("times")
    pdf.set_fill_color(240)

    pdf.multi_cell(
        fill=True,
        w=0,
        text=LOREM_IPSUM * 7,
        padding=[2, 0, 0, 0],  # top, right, bottom, left
    )

    assert pdf.page_no() > 1, "Must be more than one page to test page break"
    assert_pdf_equal(pdf, HERE / "multi_cell_page_break_with_fill.pdf", tmp_path)


def test_multi_cell_without_any_font_set():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(FPDFException) as error:
        pdf.multi_cell(text="Hello world!", w=pdf.epw)
    assert str(error.value) == "No font set, you need to call set_font() beforehand"


def test_ln_positioning_and_page_breaking_for_multicell(tmp_path):
    doc = FPDF(format="letter", unit="pt")
    doc.add_page()
    doc.set_font("helvetica", size=TEXT_SIZE)

    doc.multi_cell(
        w=144,
        h=LINE_HEIGHT,
        border=1,
        text=LOREM_IPSUM[:29],
        new_x="RIGHT",
        new_y="NEXT",
    )
    doc.multi_cell(
        w=180,
        h=LINE_HEIGHT,
        border=1,
        text=LOREM_IPSUM[29:60],
        new_x="LEFT",
        new_y="NEXT",
    )
    doc.multi_cell(
        w=144,
        h=LINE_HEIGHT,
        border=1,
        text=LOREM_IPSUM[60:90],
        new_x="LMARGIN",
        new_y="NEXT",
    )
    doc.cell(
        w=72 * 5,
        h=LINE_HEIGHT,
        border=1,
        text=LOREM_IPSUM[0:30],
        new_x="LMARGIN",
        new_y="NEXT",
    )
    doc.cell(
        w=72 * 5,
        h=LINE_HEIGHT,
        border=1,
        text=LOREM_IPSUM[31:60],
        new_x="LMARGIN",
        new_y="NEXT",
    )
    doc.cell(
        w=72 * 5,
        h=LINE_HEIGHT,
        border=1,
        text=LOREM_IPSUM[61:90],
        new_x="LMARGIN",
        new_y="NEXT",
    )
    doc.cell(
        w=72 * 5,
        h=LINE_HEIGHT,
        border=1,
        text=LOREM_IPSUM[91:120],
        new_x="LMARGIN",
        new_y="NEXT",
    )
    doc.cell(w=72 * 5, h=LINE_HEIGHT, border=1)
    doc.cell(w=1, h=LINE_HEIGHT, new_x="LEFT", new_y="NEXT")
    doc.multi_cell(
        w=144,
        h=LINE_HEIGHT,
        border=1,
        text=LOREM_IPSUM[30:90],
        new_x="LEFT",
        new_y="NEXT",
    )
    doc.cell(
        w=72 * 2,
        h=LINE_HEIGHT,
        border=1,
        text="Lorem ipsum",
        new_x="LEFT",
        new_y="NEXT",
    )
    doc.cell(
        w=72 * 2,
        h=LINE_HEIGHT,
        border=1,
        text="Lorem ipsum",
        new_x="LEFT",
        new_y="NEXT",
    )

    assert_pdf_equal(
        doc, HERE / "ln_positioning_and_page_breaking_for_multicell.pdf", tmp_path
    )


def test_multi_cell_border_thickness(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=TEXT_SIZE)
    doc.set_line_width(3)
    doc.multi_cell(w=45, h=LINE_HEIGHT, border=1, text="Lorem")
    doc.multi_cell(w=45, h=LINE_HEIGHT, border=1, text="ipsum")
    doc.multi_cell(w=45, h=LINE_HEIGHT, border=1, text="Ut")
    doc.multi_cell(w=45, h=LINE_HEIGHT, border=1, text="nostrud")
    assert_pdf_equal(doc, HERE / "multi_cell_border_thickness.pdf", tmp_path)


def test_multi_cell_ln_1(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=TEXT_SIZE)
    doc.multi_cell(
        w=100,
        h=LINE_HEIGHT,
        border=1,
        text="Lorem ipsum",
        new_x="LMARGIN",
        new_y="NEXT",
    )
    doc.multi_cell(w=100, h=LINE_HEIGHT, border=1, text="Ut nostrud irure")
    assert_pdf_equal(doc, HERE / "multi_cell_ln_1.pdf", tmp_path)


def test_multi_cell_ln_3(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=TEXT_SIZE)
    doc.multi_cell(
        w=45, h=LINE_HEIGHT, border=1, text="Lorem", new_x="RIGHT", new_y="TOP"
    )
    doc.multi_cell(
        w=45, h=LINE_HEIGHT, border=1, text="ipsum", new_x="RIGHT", new_y="TOP"
    )
    doc.multi_cell(w=45, h=LINE_HEIGHT, border=1, text="Ut", new_x="RIGHT", new_y="TOP")
    doc.multi_cell(
        w=45, h=LINE_HEIGHT, border=1, text="nostrud", new_x="RIGHT", new_y="TOP"
    )
    assert_pdf_equal(doc, HERE / "multi_cell_ln_3.pdf", tmp_path)


def test_multi_cell_ln_3_table(tmp_path):
    """
    Test rendering of a table with multi-lines cell contents
    cf. https://github.com/py-pdf/fpdf2/issues/63
    """
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=10)
    line_height = pdf.font_size * 2.5
    # Set column width to 1/4 of effective page width to distribute content
    # evenly across table and page
    col_width = pdf.epw / 4
    for row in TABLE_DATA:
        for datum in row:
            pdf.multi_cell(
                col_width,
                line_height,
                str(datum),
                border=1,
                new_x="RIGHT",
                new_y="TOP",
                max_line_height=pdf.font_size,
            )
        pdf.ln(line_height)
    assert_pdf_equal(pdf, HERE / "multi_cell_ln_3_table.pdf", tmp_path)


def test_multi_cell_table_with_automatic_page_break(tmp_path):  # issue 120
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    line_height = pdf.font_size * 2
    col_width = pdf.epw / 4  # distribute content evenly
    for _ in range(5):  # repeat table 5 times
        for row in TABLE_DATA:
            for datum in row:
                pdf.multi_cell(
                    col_width,
                    line_height,
                    datum,
                    border=1,
                    new_x="RIGHT",
                    new_y="TOP",
                    max_line_height=pdf.font_size,
                )
            pdf.ln(line_height)
    assert_pdf_equal(
        pdf, HERE / "multi_cell_table_with_automatic_page_break.pdf", tmp_path
    )


def test_multi_cell_table_with_max_line_height(tmp_path):  # issue 589
    """
    When using multi_cell() with max_line_height to render multiline text,
    the last line should be rendered like all the others
    """
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica")
    text = (
        "Discard the water and serve the boiled gaozis."
        " Tip: If you love spicy dishes, add a bit of our Red Silk Chili"
        " (not included) with the gaozis"
    )
    pdf.multi_cell(w=120, h=50, text=text, max_line_height=6, border=True)
    pdf.ln()
    pdf.multi_cell(w=120, h=18, text=text, max_line_height=6, border=True)
    assert_pdf_equal(pdf, HERE / "multi_cell_table_with_max_line_height.pdf", tmp_path)


def test_multi_cell_justified_with_unicode_font(tmp_path):  # issue 118
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(fname=FONTS_DIR / "DejaVuSans.ttf")
    pdf.set_font("DejaVuSans", size=14)
    text = 'Justified line containing "()" that is long enough to trigger wrapping and a line jump'
    pdf.multi_cell(w=0, h=8, text=text, new_x="LMARGIN", new_y="NEXT")
    assert_pdf_equal(pdf, HERE / "multi_cell_justified_with_unicode_font.pdf", tmp_path)


def test_multi_cell_split_only():  # discussion 314
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=TEXT_SIZE)
    text = "Lorem ipsum Ut nostrud irure reprehenderit anim nostrud dolore sed ut"
    expected = [
        "Lorem ipsum Ut nostrud irure",
        "reprehenderit anim nostrud",
        "dolore sed ut",
    ]
    with pytest.warns(
        DeprecationWarning, match='The parameter "split_only" is deprecated.'
    ) as record:
        assert (
            pdf.multi_cell(w=0, h=LINE_HEIGHT, text=text, split_only=True) == expected
        )
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)


def test_multi_cell_with_empty_contents(tmp_path):  # issue 349
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=10)
    for i in range(1, 5):
        pdf.multi_cell(20, new_x="RIGHT", new_y="TOP", text=str(i))
    pdf.ln(10)
    for i in range(1, 5):
        pdf.multi_cell(20, new_x="RIGHT", new_y="TOP", text=str(i) if i > 2 else "")
    assert_pdf_equal(pdf, HERE / "multi_cell_with_empty_contents.pdf", tmp_path)


def test_multicell_badinput():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pytest.raises(ValueError):
        pdf.multi_cell(0, ln=5)
    with pytest.raises(TypeError):
        pdf.multi_cell(0, new_x=5)
    with pytest.raises(TypeError):
        pdf.multi_cell(0, new_y=None)
    with pytest.raises(ValueError):
        # w as string instead of float
        pdf.multi_cell("width")
    with pytest.raises(ValueError):
        # h as string instead of float
        pdf.multi_cell(0, "height")


def test_multi_cell_j_paragraphs(tmp_path):  # issue 364
    pdf = FPDF(format="A5")
    pdf.add_page()
    pdf.add_font(fname=FONTS_DIR / "DejaVuSans.ttf")
    pdf.set_font("DejaVuSans", size=14)
    pdf.set_margins(34, 55, 34)
    pdf.set_auto_page_break(auto=True, margin=55)
    # pylint: disable=line-too-long
    text = """« Jadis, si je me souviens bien, ma vie était un festin où s’ouvraient tous les cœurs, où tous les vins coulaient.

Un soir, j’ai assis la Beauté sur mes genoux. — Et je l’ai trouvée amère. — Et je l’ai injuriée.

Je me suis armé contre la justice.

Je me suis enfui. Ô sorcières, ô misère, ô haine, c’est à vous que mon trésor a été confié !

Je parvins à faire s’évanouir dans mon esprit toute l’espérance humaine. Sur toute joie pour l’étrangler j’ai fait le bond sourd de la bête féroce.

J’ai appelé les bourreaux pour, en périssant, mordre la crosse de leurs fusils. J’ai appelé les fléaux, pour m’étouffer avec le sable, le sang. Le malheur a été mon dieu. Je me suis allongé dans la boue. Je me suis séché à l’air du crime. Et j’ai joué de bons tours à la folie."""

    pdf.multi_cell(w=0, h=None, text=text, align="J")
    assert_pdf_equal(pdf, HERE / "multi_cell_j_paragraphs.pdf", tmp_path)


def test_multi_cell_font_leakage(tmp_path):  # Issue #359
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", fname=FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", style="B", fname=FONTS_DIR / "Roboto-Bold.ttf")
    pdf.set_font("Roboto", size=12)

    pdf.multi_cell(0, text="xyz **abcde**", markdown=True)
    pdf.ln()
    pdf.set_font("Roboto", size=12)
    pdf.multi_cell(0, text="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
    pdf.ln()
    pdf.ln()
    pdf.multi_cell(0, text="xyz **abcde** ", markdown=True)
    pdf.ln()
    pdf.set_font("Roboto", size=12)
    pdf.multi_cell(0, text="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
    assert_pdf_equal(pdf, HERE / "multi_cell_font_leakage.pdf", tmp_path)


def test_multi_cell_with_zero_horizontal_space():  # issue #389
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    pdf.multi_cell(w=0, h=5, text="test")
    with pytest.raises(FPDFException):
        pdf.multi_cell(w=0, h=5, text="test")


def test_multi_cell_with_limited_horizontal_space():  # issue #389
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    pdf.multi_cell(w=pdf.epw - 2 * pdf.c_margin - 1, h=5, text="test")
    assert pdf.x == pdf.l_margin + pdf.epw - 2 * pdf.c_margin - 1
    with pytest.raises(FPDFException):
        pdf.multi_cell(w=0, h=5, text="test")


def test_multi_cell_trailing_nl(tmp_path):  # issue #455
    """Each multi_line() call triggers a line break at the end."""
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    lines = ["Hello\n", "Sweet\n", "World\n"]
    for line in lines:
        pdf.multi_cell(200, text=line)
    pdf.cell(text="end_mmc")
    pdf.ln(50)
    pdf.multi_cell(200, text="".join(lines))
    pdf.cell(text="end_mc")
    assert_pdf_equal(pdf, HERE / "multi_cell_trailing_nl.pdf", tmp_path)


def test_multi_cell_font_stretching(tmp_path):  # issue #478
    pdf = FPDF()
    pdf.add_page()
    # built-in font
    pdf.set_font("Helvetica", size=8)
    pdf.set_fill_color(255, 255, 0)
    pdf.multi_cell(w=50, text=LOREM_IPSUM[:100], new_x="LEFT", fill=True)
    pdf.ln()
    pdf.set_stretching(150)
    pdf.multi_cell(w=50, text=LOREM_IPSUM[:100], new_x="LEFT", fill=True)
    pdf.ln()
    # unicode font
    pdf.set_stretching(100)
    pdf.add_font(fname=FONTS_DIR / "DroidSansFallback.ttf")
    pdf.set_font("DroidSansFallback", size=8)
    pdf.set_fill_color(255, 255, 0)
    pdf.multi_cell(w=50, text=LOREM_IPSUM[:100], new_x="LEFT", fill=True)
    pdf.ln()
    pdf.set_stretching(150)
    pdf.multi_cell(w=50, text=LOREM_IPSUM[:100], new_x="LEFT", fill=True)
    assert_pdf_equal(pdf, HERE / "multi_cell_font_stretching.pdf", tmp_path)


def test_multi_cell_char_spacing(tmp_path):  # issue #489
    pdf = FPDF()
    pdf.add_page()
    # built-in font
    pdf.set_font("Helvetica", size=8)
    pdf.set_fill_color(255, 255, 0)
    pdf.multi_cell(w=150, text=LOREM_IPSUM[:200], new_x="LEFT", fill=True)
    pdf.ln()
    pdf.set_char_spacing(10)
    pdf.multi_cell(w=150, text=LOREM_IPSUM[:200], new_x="LEFT", fill=True)
    pdf.ln()
    # unicode font
    pdf.set_char_spacing(0)
    pdf.add_font(fname=FONTS_DIR / "DroidSansFallback.ttf")
    pdf.set_font("DroidSansFallback", size=8)
    pdf.set_fill_color(255, 255, 0)
    pdf.multi_cell(w=150, text=LOREM_IPSUM[:200], new_x="LEFT", fill=True)
    pdf.ln()
    pdf.set_char_spacing(10)
    pdf.multi_cell(w=150, text=LOREM_IPSUM[:200], new_x="LEFT", fill=True)
    assert_pdf_equal(pdf, HERE / "multi_cell_char_spacing.pdf", tmp_path)


def test_multi_cell_char_wrap(tmp_path):  # issue #649
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    pdf.set_fill_color(255, 255, 0)
    pdf.multi_cell(w=50, text=LOREM_IPSUM[:200], new_x="LEFT", fill=True)
    pdf.ln()
    pdf.multi_cell(
        w=50, text=LOREM_IPSUM[:200], new_x="LEFT", fill=True, wrapmode="CHAR"
    )
    pdf.ln()
    pdf.set_font("Courier", size=10)
    txt = "     " + "abcdefghijklmnopqrstuvwxyz" * 3
    pdf.multi_cell(w=50, text=txt, new_x="LEFT", fill=True, align="L")
    pdf.ln()
    pdf.multi_cell(w=50, text=txt, new_x="LEFT", fill=True, align="L", wrapmode="CHAR")
    assert_pdf_equal(pdf, HERE / "multi_cell_char_wrap.pdf", tmp_path)


def test_multi_cell_centering(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.multi_cell(w=120, text=LOREM_IPSUM, border=1, center=True)
    assert_pdf_equal(pdf, HERE / "multi_cell_centering.pdf", tmp_path)


def test_multi_cell_align_x(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.set_x(140)
    pdf.multi_cell(w=120, text=LOREM_IPSUM, border=1, align="X")
    pdf.set_draw_color(r=0, g=255, b=0)
    pdf.line(140, 0, 140, pdf.h)
    assert_pdf_equal(pdf, HERE / "multi_cell_align_x.pdf", tmp_path)


def test_multi_cell_centering_and_align_x(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.multi_cell(w=120, text=LOREM_IPSUM, border=1, center=True, align="X")
    pdf.set_draw_color(r=0, g=255, b=0)
    pdf.line(pdf.w / 2, 0, pdf.w / 2, pdf.h)
    assert_pdf_equal(pdf, HERE / "multi_cell_centering_and_align_x.pdf", tmp_path)


def test_multi_cell_deprecated_txt_arg():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=TEXT_SIZE)
    with pytest.warns(
        DeprecationWarning, match='The parameter "txt" has been renamed to "text"'
    ):
        # pylint: disable=unexpected-keyword-arg
        pdf.multi_cell(w=0, txt="Lorem ipsum Ut nostrud irure")


def test_multi_cell_align_with_padding(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    pdf.set_fill_color(222)

    # show alignment markers
    with pdf.local_context(draw_color=(0, 255, 0)):
        pdf.line(pdf.l_margin, 0, pdf.l_margin, pdf.h)
        pdf.line(pdf.w / 2, 0, pdf.w / 2, pdf.h)

    def create_boxes(new_x, new_y, align, padding=2):
        w = (pdf.w - pdf.l_margin - pdf.r_margin) / 6
        for ch in "ABC":
            text = f"{ch}1\n{ch}{ch}2\n{ch}{ch}{ch}3"
            pdf.multi_cell(
                w=w,
                text=text,
                fill=True,
                border=True,
                padding=padding,
                new_x=new_x,
                new_y=new_y,
                align=align,
            )

    create_boxes("LMARGIN", "NEXT", "JUSTIFY")
    create_boxes("RIGHT", "TOP", "CENTER")
    create_boxes("LEFT", "NEXT", "RIGHT")
    create_boxes("CENTER", "NEXT", "X")
    pdf.ln()
    create_boxes("START", "NEXT", "LEFT")
    create_boxes("END", "NEXT", "RIGHT")
    assert_pdf_equal(pdf, HERE / "multi_cell_align_with_padding.pdf", tmp_path)


def test_multi_cell_with_padding(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.multi_cell(0, 5, LONG_TEXT, border=1, padding=(10, 20, 30, 40))

    assert_pdf_equal(pdf, HERE / "multi_cell_with_padding.pdf", tmp_path)


def test_multi_cell_with_padding_check_input():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)

    with pytest.raises(ValueError):
        pdf.multi_cell(0, 5, LONG_TEXT, border=1, padding=(5, 5, 5, 5, 5, 5))


def test_multi_cell_return_value(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)

    pdf.x = 5

    out = pdf.multi_cell(
        0,
        5,
        "Monty Python\nKiller Sheep",
        border=1,
        padding=0,
        output=MethodReturnValue.PAGE_BREAK | MethodReturnValue.HEIGHT,
    )
    height_without_padding = out[1]

    pdf.x = 5
    # pdf.y += 50

    # try again
    out = pdf.multi_cell(
        0,
        5,
        "Monty Python\nKiller Sheep",
        border=1,
        padding=0,
        output=MethodReturnValue.PAGE_BREAK | MethodReturnValue.HEIGHT,
    )

    height_without_padding2 = out[1]

    pdf.x = 5
    # pdf.y += 50

    # try again
    out = pdf.multi_cell(
        0,
        5,
        "Monty Python\nKiller Sheep",
        border=1,
        padding=10,
        output=MethodReturnValue.PAGE_BREAK | MethodReturnValue.HEIGHT,
    )

    height_with_padding = out[1]

    assert height_without_padding == height_without_padding2
    assert height_without_padding + 20 == height_with_padding

    pdf.x = 5
    pdf.y += 10

    out = pdf.multi_cell(
        0,
        5,
        "Monty Python\nKiller Sheep",
        border=1,
        padding=10,
        output=MethodReturnValue.PAGE_BREAK | MethodReturnValue.HEIGHT,
        new_y=YPos.NEXT,
    )

    assert_pdf_equal(pdf, HERE / "multi_cell_return_value.pdf", tmp_path)


def test_multi_cell_markdown_bleeding(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=60)
    pdf.multi_cell(
        w=0, text="**Lorem Ipsum dolor**", markdown=True, new_x="LMARGIN", new_y="NEXT"
    )
    assert pdf.font_style == ""
    pdf.multi_cell(
        w=0, text="No Markdown", markdown=False, new_x="LMARGIN", new_y="NEXT"
    )
    pdf.multi_cell(
        w=0, text="--Lorem Ipsum dolor--", markdown=True, new_x="LMARGIN", new_y="NEXT"
    )
    assert pdf.font_style == ""
    pdf.multi_cell(
        w=0, text="No Markdown", markdown=False, new_x="LMARGIN", new_y="NEXT"
    )
    pdf.multi_cell(
        w=0, text="__Lorem Ipsum dolor__", markdown=True, new_x="LMARGIN", new_y="NEXT"
    )
    assert pdf.font_style == ""
    pdf.multi_cell(
        w=0, text="No Markdown", markdown=False, new_x="LMARGIN", new_y="NEXT"
    )
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_bleeding.pdf", tmp_path)



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_rect_clip(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()

    top_left_x, top_left_y = 50, 50
    with pdf.rect_clip(x=top_left_x + 8, y=top_left_y, w=46, h=32):
        pdf.image(
            HERE / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png",
            x=top_left_x,
            y=top_left_y,
        )

    top_left_x, top_left_y = 50, 150
    with pdf.rect_clip(x=top_left_x + 10, y=top_left_y + 34, w=34, h=20):
        pdf.image(
            HERE / "image_types/insert_images_insert_png.png",
            x=top_left_x,
            y=top_left_y,
        )

    assert_pdf_equal(pdf, HERE / "rect_clip.pdf", tmp_path)


def test_round_clip(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()

    top_left_x, top_left_y = 50, 50
    with pdf.round_clip(x=top_left_x + 12, y=top_left_y + 2, r=38):
        pdf.image(
            HERE / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png",
            x=top_left_x,
            y=top_left_y,
        )

    top_left_x, top_left_y = 50, 150
    with pdf.round_clip(x=top_left_x + 12, y=top_left_y + 25, r=30):
        pdf.image(
            HERE / "image_types/insert_images_insert_png.png",
            x=top_left_x,
            y=top_left_y,
        )

    assert_pdf_equal(pdf, HERE / "round_clip.pdf", tmp_path)


def test_elliptic_clip(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()

    with pdf.elliptic_clip(x=90, y=50, w=100, h=60):
        pdf.image(HERE / "png_images/d2e515cfdabae699301dcf290382474d.png", w=pdf.epw)

    assert_pdf_equal(pdf, HERE / "elliptic_clip.pdf", tmp_path)



from pathlib import Path

from fpdf import FPDF
from fpdf.unicode_script import get_unicode_script, UnicodeScript
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"


def test_indi_text(tmp_path):  # issue #365
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Mangal", fname=HERE / "Mangal 400.ttf")
    pdf.set_font("Mangal", size=40)
    pdf.set_text_shaping(False)
    pdf.cell(text="इण्टरनेट पर हिन्दी के साधन", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(True)
    pdf.cell(text="इण्टरनेट पर हिन्दी के साधन", new_x="LEFT", new_y="NEXT")

    assert_pdf_equal(pdf, HERE / "shaping_hindi.pdf", tmp_path)


def test_mixed_text_shaping(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.r_margin = 100
    pdf.add_font(
        family="KFGQPC", fname=HERE / "KFGQPC Uthmanic Script HAFS Regular.otf"
    )
    pdf.set_font("KFGQPC", size=36)
    pdf.set_text_shaping(True)
    pdf.write(text="مثال على اللغة العربية. محاذاة لليمين.")
    pdf.add_font(family="Mangal", fname=HERE / "Mangal 400.ttf")
    pdf.set_font("Mangal", size=40)
    # With preceding whitespace
    pdf.write(text=" इण्टरनेट पर हिन्दी के साधन")

    assert_pdf_equal(pdf, HERE / "text_mixed_text_shaping.pdf", tmp_path)


def test_text_replacement(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="FiraCode", fname=HERE / "FiraCode-Regular.ttf")
    pdf.set_font("FiraCode", size=40)
    pdf.set_text_shaping(False)
    pdf.cell(text="http://www 3 >= 2 != 1", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(True)
    pdf.cell(text="http://www 3 >= 2 != 1", new_x="LEFT", new_y="NEXT")

    assert_pdf_equal(pdf, HERE / "text_replacement.pdf", tmp_path)


def test_kerning(tmp_path):  # issue #812
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Dumbledor3Thin", fname=HERE / "Dumbledor3Thin.ttf")
    pdf.set_font("Dumbledor3Thin", size=40)
    pdf.set_text_shaping(False)
    pdf.cell(text="Ты То Тф Та Тт Ти", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(True)
    pdf.cell(text="Ты То Тф Та Тт Ти", new_x="LEFT", new_y="NEXT")

    assert_pdf_equal(pdf, HERE / "kerning.pdf", tmp_path)


def test_hebrew_diacritics(tmp_path):  # issue #549
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="SBL_Hbrw", fname=HERE / "SBL_Hbrw.ttf")
    pdf.set_font("SBL_Hbrw", size=40)
    pdf.set_text_shaping(False)
    pdf.cell(text="בּ", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(True)
    pdf.cell(text="בּ", new_x="LEFT", new_y="NEXT")

    assert_pdf_equal(pdf, HERE / "hebrew_diacritics.pdf", tmp_path)


def test_ligatures(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="ViaodaLibre", fname=HERE / "ViaodaLibre-Regular.ttf")
    pdf.set_font("ViaodaLibre", size=40)
    pdf.set_text_shaping(False)
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(True)
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")

    assert_pdf_equal(pdf, HERE / "ligatures.pdf", tmp_path)


def test_arabic_right_to_left(tmp_path):  # issue #549
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(
        family="KFGQPC", fname=HERE / "KFGQPC Uthmanic Script HAFS Regular.otf"
    )
    pdf.set_font("KFGQPC", size=36)
    pdf.set_text_shaping(False)
    pdf.cell(text="مثال على اللغة العربية. محاذاة لليمين.", new_x="LEFT", new_y="NEXT")
    pdf.ln(36)
    pdf.set_text_shaping(True)
    pdf.cell(text="مثال على اللغة العربية. محاذاة لليمين.", new_x="LEFT", new_y="NEXT")

    assert_pdf_equal(pdf, HERE / "arabic.pdf", tmp_path)


def test_multi_cell_markdown_with_shaping(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", "B", FONTS_DIR / "Roboto-Bold.ttf")
    pdf.add_font("Roboto", "I", FONTS_DIR / "Roboto-Italic.ttf")
    pdf.set_font("Roboto", size=32)
    pdf.set_text_shaping(True)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum dolor, **consectetur adipiscing** elit,"
        " eiusmod __tempor incididunt__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_with_styling.pdf", tmp_path)


def test_features(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="ViaodaLibre", fname=HERE / "ViaodaLibre-Regular.ttf")
    pdf.set_font("ViaodaLibre", size=40)
    pdf.set_text_shaping(use_shaping_engine=True)
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(use_shaping_engine=True, features={"liga": False})
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(use_shaping_engine=True, features={"kern": False})
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(
        use_shaping_engine=True, direction="rtl", script="Latn", language="en-us"
    )
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()

    assert_pdf_equal(pdf, HERE / "features.pdf", tmp_path)


def test_text_with_parentheses(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="SBL_Hbrw", fname=HERE / "SBL_Hbrw.ttf")
    pdf.set_font("SBL_Hbrw", size=30)
    pdf.set_text_shaping(True)
    pdf.cell(text="אנגלית (באנגלית: English) ה", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.cell(text="אנגלית (באנגלית: English) ", new_y="NEXT")
    assert_pdf_equal(pdf, HERE / "text_with_parentheses.pdf", tmp_path)


def test_text_shaping_and_offset_rendering(tmp_path):  # issue #1075
    pdf = FPDF()
    pdf.add_font("Garuda", fname=FONTS_DIR / "Garuda.ttf")
    pdf.set_font("Garuda", size=16)
    pdf.set_text_shaping(True)
    pdf.add_page()
    line_height, col_width = pdf.font_size * 2, pdf.epw / 4
    for i in range(2):
        with pdf.offset_rendering():
            pdf.cell(col_width, line_height, f"Cell ({i})")
        pdf.cell(col_width, line_height, f"Cell ({i})")
    assert_pdf_equal(pdf, HERE / "text_shaping_and_offset_rendering.pdf", tmp_path)


def test_multilingual_string(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="KhmerOS", fname=HERE / "KhmerOS.ttf")
    pdf.set_font("KhmerOS", size=30)
    pdf.set_text_shaping(True)
    pdf.cell(text="Khmer: សួស្តី ពិភពលោក")
    assert_pdf_equal(pdf, HERE / "multilingual_string.pdf", tmp_path)


def test_unicode_script_codes():
    char_list = (
        "\x03",
        "E",
        "Ͳ",
        "Ѧ",
        "Ւ",
        "ֲ",
        "\u0601",
        "܋",
        "ޔ",
        "ं",
        "ঀ",
        "ਁ",
        "ઁ",
        "ଁ",
        "ஂ",
        "ఀ",
        "ಀ",
        "ഁ",
        "ඁ",
        "พ",
        "ຂ",
        "ༀ",
        "ဦ",
        "Ⴔ",
        "ᆙ",
        "ለ",
        "Ꭴ",
        "᐀",
        "\u1680",
        "ᚲ",
        "ឨ",
        "᠀",
        "を",
        "ケ",
        "˫",
        "⺁",
        "ꀑ",
        "𐌒",
        "𐌿",
        "𐐱",
        "̬",
        "ᜃ",
        "ᜬ",
        "ᝈ",
        "ᝡ",
        "ᤌ",
        "ᥢ",
        "𐀁",
        "𐎝",
        "𐑷",
        "𐒒",
        "𐠂",
        "⣫",
        "ᨏ",
        "ϯ",
        "ᦐ",
        "Ⰺ",
        "ⵎ",
        "ꠀ",
        "𐎧",
        "𐨀",
        "ᬀ",
        "𒉑",
        "𐤀",
        "ꡪ",
        "߂",
        "ᮁ",
        "ᰙ",
        "᱘",
        "ꗵ",
        "ꢁ",
        "꤈",
        "ꤲ",
        "𐊌",
        "𐊧",
        "𐤯",
        "ꨤ",
        "ᩂ",
        "ꪑ",
        "𐬏",
        "𓆊",
        "ࠍ",
        "ꓧ",
        "ꛆ",
        "ꦁ",
        "ꫪ",
        "𐡄",
        "𐩶",
        "𐭂",
        "𐭦",
        "𐰞",
        "𑂁",
        "ᯣ",
        "𑀀",
        "ࡍ",
        "𑄂",
        "𐦪",
        "𐦈",
        "𖼴",
        "𑆁",
        "𑃝",
        "𑚛",
        "𐕘",
        "𖫝",
        "𛱨",
        "𐔑",
        "𑌁",
        "𖬫",
        "𑈇",
        "𐙤",
        "𑅙",
        "𐫀",
        "𞠒",
        "𑘑",
        "𖩍",
        "𐪁",
        "𐢜",
        "𐡰",
        "𑫘",
        "𐍒",
        "𐮅",
        "𑖤",
        "𑋒",
        "𑒆",
        "𑣅",
        "𑜄",
        "𔕘",
        "𐣥",
        "𑊅",
        "𐲑",
        "𝣣",
        "𞥂",
        "𑰈",
        "𑱰",
        "𑐩",
        "𐒵",
        "𖿠",
        "𑴄",
        "𖿡",
        "𑩐",
        "𑨀",
        "𑠇",
        "𑵢",
        "𑻧",
        "𖹻",
        "𐴞",
        "𐼾",
        "𐼅",
        "𐿴",
        "𑦤",
        "𞄥",
        "𞋄",
        "𐾷",
        "𑤁",
        "𖿤",
        "𐺟",
        "𒿬",
        "𐽷",
        "𖪼",
        "𞊞",
        "𐕸",
        "\U00011f00",
        "\U0001e4d5",
    )
    for index, char in enumerate(char_list):
        assert get_unicode_script(char) == UnicodeScript(index)
    get_unicode_script.cache_clear()


def test_disabling_text_shaping(tmp_path):  # issue #1287
    pdf = FPDF()
    pdf.add_font(family="ViaodaLibre", fname=HERE / "ViaodaLibre-Regular.ttf")
    pdf.set_font("ViaodaLibre", size=24)
    pdf.add_page()
    assert not pdf.text_shaping
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(True)
    assert pdf.text_shaping
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(False)
    assert not pdf.text_shaping
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    pdf.set_text_shaping(True)
    assert pdf.text_shaping
    pdf.cell(text="final soft stuff", new_x="LEFT", new_y="NEXT")
    pdf.ln()
    assert_pdf_equal(pdf, HERE / "disabling_text_shaping.pdf", tmp_path)



import json
import zlib
from io import BytesIO
from pathlib import Path

import fpdf

HERE = Path(__file__).resolve().parent


def break_down_filename(image):
    """
    filename:                               g04i2c08.png
                                            || ||||
    test feature (in this case gamma) ------+| ||||
    parameter of test (here gamma-value) ----+ ||||
    interlaced or non-interlaced --------------+|||
    color-type (numerical) ---------------------+||
    color-type (descriptive) --------------------+|
    bit-depth ------------------------------------|
    """
    return {
        "test_feature": image[0].lower(),
        "parameter_ot": image[1:3].lower(),
        "noninterlace": image[3].lower() != "i",
        "colortype_nm": int(image[4], 10),
        "colortype_ds": image[5].lower(),
        "n_bits_depth": int(image[6:8], 10),
    }


def test_get_img_info():
    short_keys = {"f", "h", "bpc", "w", "cs", "trns", "dp", "pal"}
    expected = json.loads((HERE / "image_info.json").read_text(encoding="utf8"))

    for path in sorted((HERE / "png_test_suite").glob("*.png")):
        if not path.stem.startswith("x"):
            blob = BytesIO(path.read_bytes())
            info = fpdf.image_parsing.get_img_info(blob)
            short_info = {}
            for k, v in info.items():
                if k == "smask":
                    short_info["smask"] = zlib.decompress(v)
                elif k in short_keys:
                    short_info[k] = v.decode("latin-1") if isinstance(v, bytes) else v

            if "smask" in expected[path.name]:
                expected[path.name]["smask"] = zlib.decompress(
                    expected[path.name]["smask"].encode("latin-1")
                )
            assert short_info == expected[path.name]


def test_get_img_info_data_rgba():
    blob = BytesIO((HERE / "png_test_suite" / "basi6a08.png").read_bytes())
    info = fpdf.image_parsing.get_img_info(blob)
    assert zlib.decompress(info["data"]) == zlib.decompress(
        b"x\x9c\xb5\xd5[\x15@P\x18D\xe198\x97\x164\xa1\tMhB\x13\x9a\xd0d"
        b"\xfc%\xf6Z\xdf<\xef\xc7\x91\xd5P\xf2\xd8P\xf2\xdcP\xf2\xdaP\xf2^Q\xf2YQ"
        b"\xf2]Q\xf2[Q\xb2\x0bJ\x9f\x0bJ\x8f\x0bJ\x973J\x873J\x9b3J\x8b3J\x93\x07"
        b"\x94\x06\xb3b#J\xbdgT\x04VT\x04vT\x04N\x94:\xdf\xa8\x08\xbc\xa8\x08\x18"
        b"\xa5\xf4\x19\xa5\xf4\x18\xa5t\x19\xa5t\x18\xa58\x04V\x1c\x02+\x0e\x01\xf5\x03\xbe\xde\n\xdb"
    )


def test_get_img_info_data_palette():
    blob = BytesIO((HERE / "png_test_suite" / "basi3p02.png").read_bytes())
    info = fpdf.image_parsing.get_img_info(blob)
    assert zlib.decompress(info["data"]) == zlib.decompress(
        b"x\x9cc`\x06\x02F `\x02\x02\x06 @\xe7c\x08\xd0@\x01\x9d\xac\xc1\xab\x00\x97."
        b"\x18\x9f\x1e\n\xf0\xb9\x1e\xc4\xa7\x87\x82\x81\x8e\x87\xd1\xf40\xa8\xd2\x03\x00d\xd4\x06\x01"
    )


def test_get_img_info_data_gray():
    blob = BytesIO((HERE / "png_test_suite" / "basi0g08.png").read_bytes())
    info = fpdf.image_parsing.get_img_info(blob)
    assert zlib.decompress(info["data"]) == zlib.decompress(
        b"x\x9cc``dbfaec\xe7\xe0\xe4\xe2\xe6\xe1\xe5\xe3\x17\x10\x14\x12\x16\x11\x15"
        b"\x13\x97\x90\x94\x92\x96\x91\x95\x93gPPTRVQUS\xd7\xd0\xd4\xd2\xd6\xd1\xd5"
        b"\xd370426153\xb7\xb0\xb4\xb2\xb6\xb1\xb5\xb3gpptrvqus\xf7\xf0\xf4\xf2\xf6"
        b"\xf1\xf5\xf3\x0f\x08\x0c\n\x0e\t\r\x0b\x8f\x88\x8c\x8a\x8e\x89\x8d\x8b"
        b"gHHLJNIMK\xcf\xc8\xcc\xca\xce\xc9\xcd\xcb/(,*.)-+\xaf\xa8\xac\xaa\xae\xa9"
        b"\xad\xabghhljnimk\xef\xe8\xec\xea\xee\xe9\xed\xeb\x9f0q\xd2\xe4)S\xa7M"
        b"\x9f1s\xd6\xec9s\xe7\xcdgX\xb0p\xd1\xe2%K\x97-_\xb1r\xd5\xea5k\xd7\xad\xdf"
        b"\xb0q\xd3\xe6-[\xb7m\xdf\xb1s\xd7\xee={\xf7\xedg8p\xf0\xd0\xe1#G\x8f\x1d?q"
        b"\xf2\xd4\xe93g\xcf\x9d\xbfp\xf1\xd2\xe5+W\xaf]\xbfq\xf3\xd6\xed;w\xef"
        b"\xddgx\xf0\xf0\xd1\xe3'O\x9f=\x7f\xf1\xf2\xd5\xeb7o\xdf\xbd\xff\xf0\xf1"
        b"\xd3\xe7/_\xbf}\xff\xf1\xf3\xd7\xef?\x7f\xff\xfdg\xf8\xf7\xf7\xcf\xef_?"
        b"\x7f|\xff\xf6\xf5\xcb\xe7O\x1f?\xbc\x7f\xf7\xf6\xcd\xebW/_<\x7f\xf6\xf4"
        b"\xc9\xe3G\x0f\x1f\xdcg\xb8w\xf7\xce\xed[7o\\\xbfv\xf5\xca\xe5K\x17/\x9c?w"
        b"\xf6\xcc\xe9S'O\x1c?v\xf4\xc8\xe1C\x07\x0f\xecg\xd8\xb7w\xcf\xee];wl\xdf"
        b"\xb6u\xcb\xe6M\x1b7\xac_\xb7v\xcd\xeaU+W,_\xb6t\xc9\xe2E\x0b\x17\xccg"
        b"\x987w\xce\xecY3gL\x9f6u\xca\xe4I\x13'\xf4\xf7\xf5\xf6twuv\xb4\xb7\xb5"
        b"\xb64756\xd43\xd4\xd5\xd6TWUV\x94\x97\x95\x96\x14\x17\x15\x16\xe4\xe7\xe5"
        b"\xe6dgef\xa4\xa7\xa5\xa6$'%&\xc43\xc4\xc5\xc6DGEF\x84\x87\x85\x86\x04\x07"
        b"\x05\x06\xf8\xfb\xf9\xfax{yz\xb8\xbb\xb9\xba8;9:\xd83\xd8\xd9\xdaX[YZ\x98"
        b"\x9b\x99\x9a\x18\x1b\x19\x1a\xe8\xeb\xe9\xeahkij\xa8\xab\xa9\xaa(+)*\xc83"
        b"\xc8\xc9\xcaHKIJ\x88\x8b\x89\x8a\x08\x0b\t\n\xf0\xf3\xf1\xf2psqr\xb0\xb3"
        b"\xb1\xb2031202\xe0O\r\n\x8a\x0c\xf8S\x83\x83#\x03\xfe\xd4\x90\x90\xc8"
        b"\x80?5442\xe0O\r\x0b\x162\xe0O\r\x07\x0e2\xe0O\r\x0f\x1e2\xe0O\r\xff\xfe2"
        b"\xe0O\r\xf7\xee2\xe0O\r\xfb\xf62\xe0O\r\xf3\xe62\xe0O\ru\xb5\x0c\xf8SC\\,"
        b"\x03\xfe\xd4`g\xcb\x80?5\xc8\xc92\xe0O\rL\xcc\x00\x17\xb3\xfc\x18"
    )



#!/usr/bin/env python

from fpdf import FPDF

pdf = FPDF()
pdf.add_page()

# Add a DejaVu Unicode font (uses UTF-8)
# Supports more than 200 languages. For a coverage status see:
# http://dejavu.svn.sourceforge.net/viewvc/dejavu/trunk/dejavu-fonts/langcover.txt
pdf.add_font(fname="DejaVuSansCondensed.ttf")
pdf.set_font("DejaVuSansCondensed", size=14)

text = """
English: Hello World
Greek: Γειά σου κόσμος
Polish: Witaj świecie
Portuguese: Olá mundo
Russian: Здравствуй, Мир
Vietnamese: Xin chào thế giới
Arabic: مرحبا العالم
Hebrew: שלום עולם
"""

for txt in text.split("\n"):
    pdf.write(8, txt)
    pdf.ln(8)

# Add a Indic Unicode font (uses UTF-8)
# Supports: Bengali, Devanagari, Gujarati,
#           Gurmukhi (including the variants for Punjabi)
#           Kannada, Malayalam, Oriya, Tamil, Telugu, Tibetan
pdf.add_font(fname="gargi.ttf")
pdf.set_font("gargi", size=14)
pdf.write(8, "Hindi: नमस्ते दुनिया")
pdf.ln(20)

# Add a AR PL New Sung Unicode font (uses UTF-8)
# The Open Source Chinese Font (also supports other east Asian languages)
pdf.add_font(fname="fireflysung.ttf")
pdf.set_font("fireflysung", size=14)
pdf.write(8, "Chinese: 你好世界\n")
pdf.write(8, "Japanese: こんにちは世界\n")
pdf.ln(10)

# Add a Alee Unicode font (uses UTF-8)
# General purpose Hangul truetype fonts that contain Korean syllable
# and Latin9 (iso8859-15) characters.
pdf.add_font(fname="Eunjin.ttf")
pdf.set_font("Eunjin", size=14)
pdf.write(8, "Korean: 안녕하세요")
pdf.ln(20)

# Add a Fonts-TLWG (formerly ThaiFonts-Scalable) (uses UTF-8)
pdf.add_font(fname="Waree.ttf")
pdf.set_font("Waree", size=14)
pdf.write(8, "Thai: สวัสดีชาวโลก")
pdf.ln(20)

# Select a standard font (uses windows-1252)
pdf.set_font("helvetica", size=14)
pdf.ln(10)
pdf.write(5, "This is standard built-in font")

fn = "unicode.pdf"
pdf.output(fn)



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_custom_file_id(tmp_path):
    class PDF(FPDF):
        def file_id(self):
            return "<DEADBEEF><DEADBEEF>"

    pdf = PDF()
    assert_pdf_equal(pdf, HERE / "custom_file_id.pdf", tmp_path)


def test_no_file_id(tmp_path):
    class PDF(FPDF):
        def file_id(self):
            return None

    pdf = PDF()
    assert_pdf_equal(pdf, HERE / "no_file_id.pdf", tmp_path)



import math
from fpdf import FPDF
from fpdf.drawing_primitives import Transform
from pathlib import Path
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_transform_isolation(tmp_path):
    """
    Verifies that the transformation is strictly isolated to the 'with' block
    and does not leak to subsequent drawings.
    """
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)

    w, h = 30, 20
    y_pos = 50

    pdf.set_draw_color(0, 0, 0)  # Black
    pdf.set_line_width(0.5)
    pdf.text(20, y_pos - 5, "1. Before (Normal)")
    pdf.rect(20, y_pos, w, h)

    x_center = 80

    cx, cy = x_center + w / 2, y_pos + h / 2

    t_complex = (
        Transform.translation(cx, cy)
        @ Transform.rotation(math.radians(20))
        @ Transform.shearing(x=math.radians(10), y=0)
        @ Transform.translation(-cx, -cy)
    )

    pdf.set_draw_color(220, 50, 50)  # Red

    with pdf.transform(t_complex):
        pdf.text(x_center, y_pos - 5, "2. Inside (Transformed)")
        pdf.rect(x_center, y_pos, w, h)
    pdf.set_draw_color(0, 0, 0)
    x_right = 140

    pdf.text(x_right, y_pos - 5, "3. After (Normal)")
    pdf.rect(x_right, y_pos, w, h)

    assert_pdf_equal(pdf, HERE / "transform.pdf", tmp_path)


def test_transform_tilted_title(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)

    pdf.text(20, 20, "Normal title")

    pdf.set_draw_color(200, 200, 200)
    pdf.set_line_width(0.2)
    pdf.line(15, 30, 180, 30)

    title_anchor_x, title_anchor_y = 90, 30
    title_transform = Transform.rotation_d(12).about(title_anchor_x, title_anchor_y)
    with pdf.transform(title_transform):
        pdf.set_font("Times", style="B", size=20)
        pdf.set_text_color(30, 30, 120)
        pdf.text(60, 30, "Tilted Title")

    pdf.set_text_color(0, 0, 0)
    pdf.set_font("Helvetica", size=9)
    pdf.text(20, 45, "Baseline at y=30")

    assert_pdf_equal(pdf, HERE / "transform_title.pdf", tmp_path)


def test_transform_mirrored_text(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=11)

    pdf.text(20, 50, "Left block")
    lines = ["Mirror me", "Line two", "Line three"]
    y_pos = 60
    for line in lines:
        pdf.text(20, y_pos, line)
        y_pos += 6

    axis_x = 105
    pdf.set_draw_color(160, 0, 0)
    pdf.set_line_width(0.2)
    pdf.line(axis_x, 40, axis_x, 110)
    pdf.set_text_color(120, 0, 0)
    pdf.text(axis_x + 2, 45, "mirror axis")

    mirror = Transform.scaling(-1, 1).about(axis_x, 0)
    with pdf.transform(mirror):
        pdf.set_text_color(0, 0, 120)
        pdf.text(20, 50, "Right block (mirrored)")
        y_pos = 60
        for line in lines:
            pdf.text(20, y_pos, line)
            y_pos += 6

    pdf.set_text_color(0, 0, 0)

    assert_pdf_equal(pdf, HERE / "transform_mirror.pdf", tmp_path)



#!/usr/bin/env python3
# Script Dependencies:
#    odfpy
# USAGE: ./ods2table.py color_srgb.ods
import sys
from fpdf import FPDF, FontFace
from fpdf.drawing_primitives import color_from_hex_string
from odf.opendocument import load
from odf.table import Table, TableCell, TableRow

pdf = FPDF()
pdf.add_page()
pdf.set_font("Times", size=22)
ods = load(sys.argv[1])
for sheet in ods.getElementsByType(Table):
    with pdf.table() as table:
        for i, row in enumerate(sheet.getElementsByType(TableRow)):
            row = [str(cell) for cell in row.getElementsByType(TableCell)]
            style = None
            if i > 0:
                # We color the row based on the hexadecimal code in the 2nd column:
                style = FontFace(fill_color=color_from_hex_string(row[1]))
            table.row(row, style=style)
pdf.output("from-ods.pdf")



from fpdf.fonts import FontFace


def test_combine_fontface():
    font1 = FontFace(family="helvetica", size_pt=12)
    # providing None override should return the default style
    assert FontFace.combine(font1, override_style=None) == font1
    # overriding a None style should return the override
    assert FontFace.combine(None, font1) == font1
    font2 = FontFace(size_pt=14)
    combined = FontFace.combine(font1, font2)
    assert isinstance(combined, FontFace)
    assert combined.family == "helvetica"  # wasn't overridden
    assert combined.emphasis is None  # wasn't specified by either
    assert combined.size_pt == 14  # was overridden



# pylint: disable=import-outside-toplevel
from contextlib import contextmanager
from datetime import datetime, timezone
from time import perf_counter
from timeit import timeit
from tracemalloc import get_traced_memory, is_tracing
from types import SimpleNamespace
import functools
import gc
import hashlib
import linecache
import logging
import os
import pathlib
import shutil
import sys
import tracemalloc
from typing import Optional
import warnings

from subprocess import check_output, CalledProcessError, PIPE

import pytest

from fpdf.template import Template

QPDF_AVAILABLE = bool(shutil.which("qpdf"))
if not QPDF_AVAILABLE:
    warnings.warn(
        "qpdf command not available on the $PATH, falling back to hash-based "
        "comparisons in tests"
    )

EPOCH = datetime(1969, 12, 31, 19, 00, 00).replace(tzinfo=timezone.utc)

LOREM_IPSUM = (
    "Lorem ipsum Ut nostrud irure reprehenderit anim nostrud dolore sed "
    "ut Excepteur dolore ut sunt irure consectetur tempor eu tempor "
    "nostrud dolore sint exercitation aliquip velit ullamco esse dolore "
    "mollit ea sed voluptate commodo amet eiusmod incididunt Excepteur "
    "Excepteur officia est ea dolore sed id in cillum incididunt quis ex "
    "id aliqua ullamco reprehenderit cupidatat in quis pariatur ex et "
    "veniam consectetur et minim minim nulla ea in quis Ut in "
    "consectetur cillum aliquip pariatur qui quis sint reprehenderit "
    "anim incididunt laborum dolor dolor est dolor fugiat ut officia do "
    "dolore deserunt nulla voluptate officia mollit elit consequat ad "
    "aliquip non nulla dolor nisi magna consectetur anim sint officia "
    "sit tempor anim do laboris ea culpa eu veniam sed cupidatat in anim "
    "fugiat culpa enim Ut cillum in exercitation magna nostrud aute "
    "proident laboris est ullamco nulla occaecat nulla proident "
    "consequat in ut labore non sit id cillum ut ea quis est ut dolore "
    "nisi aliquip aute pariatur ullamco ut cillum Duis nisi elit sit "
    "cupidatat do Ut aliqua irure sunt sunt proident sit aliqua in "
    "dolore Ut in sint sunt exercitation aliquip elit velit dolor nisi "
)

EMOJI_TEST_TEXT = (
    "Remember the days when we had to rely on simple text-based emoticons like :-) or :D to show our feelings online? Those little "
    "symbols paved the way for a whole universe of expressive visuals 🤯. As technology advanced 🚀 and mobile devices became more "
    "powerful 📱, the emoticon evolved into the now-iconic emoji 🚦😍! Suddenly, instead of typing <3 for love, we could send an "
    "actual heart ❤️—and a million other icons, too. From smiling faces 😊 to dancing humans 💃, from tiny pizzas 🍕 to entire flags "
    "🌍, emojis quickly took over every conversation! Now, we can convey jokes 🤪, excitement 🤩, or even complicated feelings 🤔 with "
    "just a tap or two. Looking back, who knew those humble :-P and ;-) would evolve into the expressive rainbow of emojis 🌈 that color "
    "our digital world today?"
)

# default block size from src/libImaging/Storage.c:
PIL_MEM_BLOCK_SIZE_IN_MIB = 16


def assert_same_file(file1, file2):
    """
    Assert that two files are the same, by comparing their absolute paths.
    """
    assert os.path.normcase(os.path.abspath(file1)) == os.path.normcase(
        os.path.abspath(file2)
    )


def assert_pdf_equal(
    actual,
    expected,
    tmp_path,
    linearize=False,
    at_epoch=True,
    generate=False,
    ignore_id_changes=False,
    ignore_original_obj_ids=False,
    ignore_xref_offsets=False,
):
    """
    This compare the output of a `FPDF` instance (or `Template` instance),
    with the provided PDF file.

    The `CreationDate` of the newly generated PDF is fixed, so that it never triggers
    a diff.

    If the `qpdf` command is available on the `$PATH`, it will be used to perform the
    comparison, as it greatly helps debugging diffs. Otherwise, a hash-based comparison
    logic is used as a fallback.

    Args:
        actual: instance of `FPDF` or `Template`. The `output` or `render` method
          will be called on it.
        expected: instance of `FPDF`, `bytearray` or file path to a PDF file
          matching the expected output
        tmp_path (Path): temporary directory provided by pytest individually to the
          caller test function
        generate (bool): only generate `pdf` output to `rel_expected_pdf_filepath`
          and return. Useful to create new tests.
    """
    if isinstance(actual, Template):
        actual.render()
        actual_pdf = actual.pdf
    else:
        actual_pdf = actual
    if at_epoch:
        actual_pdf.creation_date = EPOCH
    if generate:
        assert isinstance(expected, pathlib.Path), (
            "When passing `True` to `generate`"
            "a pathlib.Path must be provided as the `expected` parameter"
        )
        actual_pdf.output(expected.open("wb"), linearize=linearize)
        return
    # Force ignore_id_changes on Python 3.14
    # CPython replaced zlib by zlib-ng on the Windows build
    # and the compressed data is not 100% identical anymore.
    # https://github.com/python/cpython/pull/131438
    if sys.version_info[:2] == (3, 14) and QPDF_AVAILABLE:
        ignore_id_changes = True
    if isinstance(expected, pathlib.Path):
        expected_pdf_path = expected
    else:
        expected_pdf_path = tmp_path / "expected.pdf"
        with expected_pdf_path.open("wb") as pdf_file:
            if isinstance(expected, (bytes, bytearray)):
                pdf_file.write(expected)
            else:
                expected.set_creation_date(EPOCH)
                expected.output(pdf_file, linearize=linearize)
    actual_pdf_path = tmp_path / "actual.pdf"
    with actual_pdf_path.open("wb") as pdf_file:
        actual_pdf.output(pdf_file, linearize=linearize)
    if QPDF_AVAILABLE:  # Favor qpdf-based comparison, as it helps a lot debugging:
        actual_qpdf = _qpdf(actual_pdf_path)
        expected_qpdf = _qpdf(expected_pdf_path)
        (tmp_path / "actual_qpdf.pdf").write_bytes(actual_qpdf)
        (tmp_path / "expected_qpdf.pdf").write_bytes(expected_qpdf)
        actual_lines = actual_qpdf.splitlines()
        expected_lines = expected_qpdf.splitlines()
        if ignore_id_changes:
            actual_lines = filter_out_doc_id(actual_lines)
            expected_lines = filter_out_doc_id(expected_lines)
        if ignore_original_obj_ids:
            actual_lines = filter_out_original_obj_ids(actual_lines)
            expected_lines = filter_out_original_obj_ids(expected_lines)
        if ignore_xref_offsets:
            actual_lines = filter_out_xref_offsets(actual_lines)
            expected_lines = filter_out_xref_offsets(expected_lines)
        if actual_lines != expected_lines:
            # It is important to reduce the size of both list of bytes here,
            # to avoid .assertSequenceEqual to take forever to finish, that itself calls difflib.ndiff,
            # that has cubic complexity from this comment by Tim Peters: https://bugs.python.org/issue6931#msg223459
            actual_lines = subst_streams_with_hashes(actual_lines)
            expected_lines = subst_streams_with_hashes(expected_lines)
        assert actual_lines == expected_lines
        if linearize:
            _run_cmd("qpdf", "--check-linearization", str(actual_pdf_path))
    else:  # Fallback to hash comparison
        actual_hash = hashlib.md5(actual_pdf_path.read_bytes()).hexdigest()
        expected_hash = hashlib.md5(expected_pdf_path.read_bytes()).hexdigest()

        assert actual_hash == expected_hash, f"{actual_hash} != {expected_hash}"


def filter_out_doc_id(lines):
    return [line for line in lines if not line.startswith(b"  /ID [<")]


def filter_out_original_obj_ids(lines):
    return [line for line in lines if not line.startswith(b"%% Original object ID: ")]


def filter_out_xref_offsets(lines):
    return [line for line in lines if not line.endswith(b" 00000 n ")]


def check_signature(pdf, trusted_cert_paths, linearize=False):
    # pylint: disable=import-outside-toplevel
    from endesive import pdf as endesive_pdf

    trusted_certs = []
    for cert_filepath in trusted_cert_paths:
        with open(cert_filepath, "rb") as cert_file:
            trusted_certs.append(cert_file.read())
    results = endesive_pdf.verify(pdf.output(linearize=linearize), trusted_certs)
    for hash_ok, signature_ok, cert_ok in results:
        assert signature_ok
        assert hash_ok
        assert cert_ok


def subst_streams_with_hashes(in_lines):
    """
    This utility function reduce the length of `in_lines`, a list of bytes,
    by replacing multi-lines streams looking like this:

        stream
        {non-printable-binary-data}endstream

    by a single line with this format:

        <stream with MD5 hash: abcdef0123456789>
    """
    out_lines, stream = [], None
    for line in in_lines:
        if line == b"stream":
            assert stream is None
            stream = bytearray()
        elif stream == b"stream":
            # First line of stream, we check if it is binary or not:
            try:
                line.decode("latin-1")
                if not (b"\0" in line or b"\xff" in line):
                    # It's likely to be text! No need to compact stream
                    stream = None
            except UnicodeDecodeError:
                pass
        if stream is None:
            out_lines.append(line)
        else:
            stream += line
        if line.endswith(b"endstream") and stream:
            stream_hash = hashlib.md5(stream).hexdigest()
            out_lines.append(f"<stream with MD5 hash: {stream_hash}>\n".encode())
            stream = None
    return out_lines


def _qpdf(input_pdf_filepath):
    return _run_cmd(
        "qpdf",
        "--deterministic-id",
        "--password=fpdf2",
        "--qdf",
        str(input_pdf_filepath),
        "-",
    )


def _run_cmd(*args):
    try:
        return check_output(args, stderr=PIPE)
    except CalledProcessError as error:
        print(f"\nqpdf STDERR: {error.stderr.decode().strip()}")
        raise


@contextmanager
def time_execution():
    """
    Usage:

        with time_execution() as duration:
            ...
        assert duration.seconds < 10
    """
    ctx = SimpleNamespace()
    start = perf_counter()
    yield ctx
    ctx.seconds = perf_counter() - start


def ensure_exec_time_below(seconds, repeats=10):
    """
    Unit test decorator using the standard timeit module
    to check that average duration of the target test
    does not get over the limit provided.

    Those checks are only enabled if $CHECK_EXEC_TIME is set.

    This decorator replaced pytest-timeout, and is a better fit:
    * because it allows to know how much above the threshold the test ran
    * because it does not cause a global PyTest interruption
    * because it computes an average, and is hence more stable
    """

    def actual_decorator(test_func):
        @functools.wraps(test_func)
        def wrapper(*args, **kwargs):
            def func_with_args():
                test_func(*args, **kwargs)

            if not os.environ.get("CHECK_EXEC_TIME"):
                func_with_args()
                return

            total_elapsed_in_secs = timeit(func_with_args, number=repeats)
            # Compute average:
            avg_duration_in_secs = total_elapsed_in_secs / repeats
            assert avg_duration_in_secs < seconds

        return wrapper

    return actual_decorator


def ensure_rss_memory_below(mib):
    """
    Ensure there is no unexpected / significant increase between
    the process RSS memory BEFORE executing the test, and AFTER.

    Those checks are only enabled if $CHECK_RSS_MEMORY is set.

    This decorator replaced memunit, and is a better fit:
    * because it takes in consideration a difference of RSS values,
      not an absolute memory amount, and hence better checks
      the memory usage of a single test, with more isolation to other tests
    * because it does not suffer from some memory_profiler issues:
        + https://github.com/py-pdf/fpdf2/issues/641#issuecomment-1465730884
        + hanging MemTimer child process sometimes preventing PyTest finalization,
          blocking in multiprocessing.util._exit_function() :
          https://github.com/python/cpython/blob/3.11/Lib/multiprocessing/util.py#L355

    Sadly, we cannot use the same approach as ensure_exec_time_below()
    of using averages, as RSS measures are made on the same process and are not independent.
    """

    def actual_decorator(test_func):
        @functools.wraps(test_func)
        def wrapper(*args, **kwargs):
            if not os.environ.get("CHECK_RSS_MEMORY"):
                test_func(*args, **kwargs)
                return
            start_rss_in_mib = get_process_rss_as_mib()
            test_func(*args, **kwargs)
            if not start_rss_in_mib:
                return  # not available under Windows
            end_rss_in_mib = get_process_rss_as_mib()
            assert end_rss_in_mib - start_rss_in_mib < mib

        return wrapper

    return actual_decorator


# Enabling this check creates an increase in memory usage,
# so we require an opt-in through a CLI argument:
def pytest_addoption(parser):
    parser.addoption(
        "--trace-memory-usage",
        action="store_true",
        help="Trace the memory usage during tests execution",
    )
    parser.addoption(
        "--pympler-summary",
        action="store_true",
        help="At the end of the tests execution, display a summary of objects allocated in memory",
    )
    parser.addoption(
        "--trace-malloc",
        action="store_true",
        help="Trace main memory allocations differences during the whole execution",
    )


def pytest_configure():
    "Disable some loggers."
    logging.getLogger("fpdf.svg").propagate = False


@pytest.fixture(scope="module", autouse=True)
def module_memory_usage(request):
    start_rss_in_mib = None
    if request.config.getoption("trace_memory_usage"):
        start_rss_in_mib = get_process_rss_as_mib()
    yield
    if not start_rss_in_mib:
        return  # not available under Windows
    gc.collect()
    end_rss_in_mib = get_process_rss_as_mib()
    sign = "+" if end_rss_in_mib > start_rss_in_mib else ""
    capmanager = request.config.pluginmanager.getplugin("capturemanager")
    with capmanager.global_and_fixture_disabled():
        print("\n")
        print(
            f"Memory bump for {request.node.name.split('/')[-1]}: {sign}{end_rss_in_mib-start_rss_in_mib:.1f} MiB"
        )


@pytest.fixture(scope="session", autouse=True)
def final_memory_usage(request):
    yield
    if request.config.getoption("trace_memory_usage"):
        gc.collect()
        capmanager = request.config.pluginmanager.getplugin("capturemanager")
        with capmanager.global_and_fixture_disabled():
            print("\n")
            print_mem_usage("Final memory usage:")


@pytest.fixture(scope="session", autouse=True)
def pympler_summary(request):
    yield
    if request.config.getoption("pympler_summary"):
        # pylint: disable=import-error
        from pympler.muppy import get_objects
        from pympler.summary import print_, summarize

        gc.collect()
        all_objects = get_objects()
        capmanager = request.config.pluginmanager.getplugin("capturemanager")
        with capmanager.global_and_fixture_disabled():
            print("\n[pympler/muppy] biggest objects summary:")
            print_(summarize(all_objects))


@pytest.fixture(scope="session", autouse=True)
def trace_malloc(request):
    if not request.config.getoption("trace_malloc"):
        yield
        return
    gc.collect()
    # Top-10 recipe from: https://docs.python.org/3/library/tracemalloc.html#display-the-top-10
    tracemalloc.start()
    snapshot1 = tracemalloc.take_snapshot().filter_traces(
        (
            tracemalloc.Filter(False, linecache.__file__),
            tracemalloc.Filter(False, tracemalloc.__file__),
        )
    )
    yield
    gc.collect()
    snapshot2 = tracemalloc.take_snapshot().filter_traces(
        (
            tracemalloc.Filter(False, linecache.__file__),
            tracemalloc.Filter(False, tracemalloc.__file__),
        )
    )
    top_stats = snapshot2.compare_to(snapshot1, "lineno")
    capmanager = request.config.pluginmanager.getplugin("capturemanager")
    with capmanager.global_and_fixture_disabled():
        print("[tracemalloc] Top 10 differences:")
        for stat in top_stats[:10]:
            print(stat)


################################################################################
################### Utility functions to track memory usage ####################
################################################################################


def print_mem_usage(prefix: str) -> None:
    print(get_mem_usage(prefix))


def get_mem_usage(prefix: str) -> str:
    _collected_count = gc.collect()
    rss = get_process_rss()
    # heap_size, stack_size = get_process_heap_and_stack_sizes()
    # objs_size_sum = get_gc_managed_objs_total_size()
    pillow = get_pillow_allocated_memory()
    # malloc_stats = "Malloc stats: " + get_pymalloc_allocated_over_total_size()
    malloc_stats = ""
    if is_tracing():
        malloc_stats = "Malloc stats: " + get_tracemalloc_traced_memory()
    return f"{prefix:<40} {malloc_stats} | Pillow: {pillow} | Process RSS: {rss}"


def get_process_rss() -> str:
    rss_as_mib = get_process_rss_as_mib()
    if rss_as_mib:
        return f"{rss_as_mib:.1f} MiB"
    return "<unavailable>"


def get_process_rss_as_mib() -> Optional[float]:
    "Inspired by psutil source code"
    pid = os.getpid()
    try:
        with open(f"/proc/{pid}/statm", encoding="utf8") as statm:
            return (
                int(statm.readline().split()[1])
                * os.sysconf("SC_PAGE_SIZE")
                / 1024
                / 1024
            )
    except FileNotFoundError:  # /proc files only exist under Linux
        return None


def get_process_heap_and_stack_sizes() -> tuple[str, str]:
    heap_size_in_mib, stack_size_in_mib = "<unavailable>", "<unavailable>"
    pid = os.getpid()
    try:
        with open(f"/proc/{pid}/maps", encoding="utf8") as maps_file:
            maps_lines = list(maps_file)
    except FileNotFoundError:  # This file only exists under Linux
        return heap_size_in_mib, stack_size_in_mib
    for line in maps_lines:
        words = line.split()
        addr_range, path = words[0], words[-1]
        addr_start_str, addr_end_str = addr_range.split("-")
        addr_start, addr_end = int(addr_start_str, 16), int(addr_end_str, 16)
        size = addr_end - addr_start
        if path == "[heap]":
            heap_size_in_mib = f"{size / 1024 / 1024:.1f} MiB"
        elif path == "[stack]":
            stack_size_in_mib = f"{size / 1024 / 1024:.1f} MiB"
    return heap_size_in_mib, stack_size_in_mib


def get_pymalloc_allocated_over_total_size() -> str:
    """
    Get PyMalloc stats from sys._debugmallocstats()
    From experiments, not very reliable
    """
    try:
        # pylint: disable=import-outside-toplevel
        from pymemtrace.debug_malloc_stats import get_debugmallocstats

        allocated, total = -1, -1
        for line in get_debugmallocstats().decode().splitlines():
            if line.startswith("Total"):
                total = int(line.split()[-1].replace(",", ""))
            elif line.startswith("# bytes in allocated blocks"):
                allocated = int(line.split()[-1].replace(",", ""))
        return f"{allocated / 1024 / 1024:.1f} / {total / 1024 / 1024:.1f} MiB"
    except ImportError:
        warnings.warn("pymemtrace could not be imported - Run: pip install pymemtrace")
        return "<unavailable>"


def get_gc_managed_objs_total_size() -> str:
    "From experiments, not very reliable"
    try:
        # pylint: disable=import-outside-toplevel
        from pympler.muppy import get_objects, getsizeof

        objs_total_size = sum(getsizeof(obj) for obj in get_objects())
        return f"{objs_total_size / 1024 / 1024:.1f} MiB"
    except ImportError:
        warnings.warn("pympler could not be imported - Run: pip install pympler")
        return "<unavailable>"


def get_tracemalloc_traced_memory() -> str:
    "Requires python -X tracemalloc"
    current, peak = get_traced_memory()
    return f"{current / 1024 / 1024:.1f} (peak={peak / 1024 / 1024:.1f}) MiB"


def get_pillow_allocated_memory() -> str:
    # pylint: disable=c-extension-no-member,import-outside-toplevel
    from PIL import Image

    stats = Image.core.get_stats()  # type: ignore[attr-defined]
    blocks_in_use = stats["allocated_blocks"] - stats["freed_blocks"]
    return f"{blocks_in_use * PIL_MEM_BLOCK_SIZE_IN_MIB:.1f} MiB"



from pathlib import Path

from fpdf import FPDF
from fpdf import FPDF_VERSION

DIR = Path(__file__).parent
FONT_DIR = DIR / ".." / "test" / "fonts"

pdf = FPDF(enforce_compliance="PDF/A-3B")
pdf.set_lang("en-US")
pdf.set_title("Tutorial7")
pdf.set_author(["John Dow", "Jane Dow"])
pdf.set_subject("Example for PDF/A")
pdf.set_keywords(["example", "tutorial", "fpdf", "pdf/a"])
pdf.set_producer(f"py-pdf/fpdf2 {FPDF_VERSION}")
pdf.add_font(fname=FONT_DIR / "DejaVuSans.ttf")
pdf.add_font("DejaVuSans", style="B", fname=FONT_DIR / "DejaVuSans-Bold.ttf")
pdf.add_font("DejaVuSans", style="I", fname=FONT_DIR / "DejaVuSans-Oblique.ttf")
pdf.add_page()
pdf.set_font("DejaVuSans", style="B", size=20)
pdf.write(text="Header")
pdf.ln(20)
pdf.set_font(size=12)
pdf.write(text="Example text")
pdf.ln(20)
pdf.set_font(style="I")
pdf.write(text="Example text in italics")

pdf.output("tuto7.pdf")



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent

RADIUS = 25
MARGIN = 10


def next_row(pdf):
    pdf.ln()
    pdf.set_xy(pdf.l_margin + RADIUS, pdf.y + 2 * RADIUS + MARGIN)


def test_circle_style(tmp_path):
    pdf = FPDF(unit="mm")
    pdf.add_page()

    pdf.set_xy(pdf.l_margin + RADIUS, pdf.y + RADIUS)
    for counter, style in enumerate(["", "F", "FD", "DF", None]):
        pdf.circle(x=pdf.x, y=pdf.y, radius=RADIUS, style=style)
        pdf.set_x(pdf.x + 2 * RADIUS + MARGIN)
        if counter % 3 == 2:
            next_row(pdf)

    assert_pdf_equal(pdf, HERE / "circle_style.pdf", tmp_path)


def test_circle_line_width(tmp_path):
    pdf = FPDF(unit="mm")
    pdf.add_page()

    pdf.set_xy(pdf.l_margin + RADIUS, pdf.y + RADIUS)
    for line_width in [1, 2, 3]:
        pdf.set_line_width(line_width)
        pdf.circle(x=pdf.x, y=pdf.y, radius=RADIUS, style=None)
        pdf.set_x(pdf.x + 2 * RADIUS + MARGIN)
    next_row(pdf)
    for line_width in [4, 5, 6]:
        pdf.set_line_width(line_width)
        pdf.circle(x=pdf.x, y=pdf.y, radius=RADIUS, style=None)
        pdf.set_x(pdf.x + 2 * RADIUS + MARGIN)
    pdf.set_line_width(0.2)  # reset

    assert_pdf_equal(pdf, HERE / "circle_line_width.pdf", tmp_path)


def test_circle_draw_color(tmp_path):
    pdf = FPDF(unit="mm")
    pdf.add_page()

    pdf.set_line_width(0.5)
    pdf.set_xy(pdf.l_margin + RADIUS, pdf.y + RADIUS)
    for gray in [70, 140, 210]:
        pdf.set_draw_color(gray)
        pdf.circle(x=pdf.x, y=pdf.y, radius=RADIUS, style=None)
        pdf.set_x(pdf.x + 2 * RADIUS + MARGIN)

    assert_pdf_equal(pdf, HERE / "circle_draw_color.pdf", tmp_path)


def test_circle_fill_color(tmp_path):
    pdf = FPDF(unit="mm")
    pdf.add_page()

    pdf.set_fill_color(240)
    pdf.set_xy(pdf.l_margin + RADIUS, pdf.y + RADIUS)
    for color in [[230, 30, 180], [30, 180, 30], [30, 30, 70]]:
        pdf.set_draw_color(*color)
        pdf.circle(x=pdf.x, y=pdf.y, radius=RADIUS, style="FD")
        pdf.set_x(pdf.x + 2 * RADIUS + MARGIN)
    next_row(pdf)

    assert_pdf_equal(pdf, HERE / "circle_fill_color.pdf", tmp_path)



"""
Tests for variable fonts feature
"""

from pathlib import Path

import pytest

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_font_without_variations(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto Variable", "", HERE / "Roboto-Variable.ttf")
    pdf.set_font("Roboto Variable", "")
    pdf.cell(
        w=pdf.epw,
        text="Roboto Variable Font - unspecified axes",
        new_x="LMARGIN",
        new_y="NEXT",
    )
    assert_pdf_equal(pdf, HERE / "variable_font_without_variations.pdf", tmp_path)


def test_font_with_variations(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(
        "Roboto Variable",
        "",
        HERE / "Roboto-Variable.ttf",
        variations={
            "B": {"wght": 900, "wdth": 75},
        },
    )
    pdf.set_font("Roboto Variable", "B", 20)
    pdf.cell(
        w=pdf.epw,
        text="Roboto Variable Font - Condensed Bold (wght=900, wdth=75)",
        new_x="LMARGIN",
        new_y="NEXT",
    )
    assert_pdf_equal(pdf, HERE / "variable_font_with_variations.pdf", tmp_path)


def test_invalid_variable_font():
    static_font_path = HERE / "Roboto-Regular.ttf"

    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(AttributeError) as error:
        pdf.add_font(
            "Roboto Variable",
            "",
            static_font_path,
            variations={
                "": {"B": {"wght": 900, "wdth": 75}},
            },
        )
        assert str(error.value) == f"{static_font_path} is not a variable font"


def test_font_already_added():
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(
        "Roboto Variable",
        "",
        HERE / "Roboto-Variable.ttf",
        variations={"": {"wght": 400, "wdth": 100}, "B": {"wght": 600, "wdth": 100}},
    )

    #  Notice the B after 'roboto variable' in the match message
    with pytest.warns(
        UserWarning,
        match="Core font or font already added 'roboto variableB': doing nothing",
    ):
        # Adding another variation of the same font with a different font weight.
        pdf.add_font(
            "Roboto Variable",
            "",
            HERE / "Roboto-Variable.ttf",
            variations={"B": {"wght": 900, "wdth": 100}},
        )


def test_font_with_one_axis(tmp_path):
    pdf = FPDF()
    pdf.add_page()

    pdf.add_font(
        "Roboto Variable",
        "",
        HERE / "Roboto-Variable.ttf",
        variations={"": {"wght": 100}},
    )
    pdf.set_font("Roboto Variable", "", 20)
    pdf.cell(
        w=pdf.epw,
        text="Roboto Variable Thin (wght=100)",
        new_x="LMARGIN",
        new_y="NEXT",
    )

    pdf.add_font(
        "Roboto Variable",
        "",
        HERE / "Roboto-Variable.ttf",
        variations={"B": {"wdth": None, "wght": 500}},
    )
    pdf.set_font("Roboto Variable", "B", 20)
    pdf.cell(
        w=pdf.epw,
        text="Roboto Variable Bolder(wdth=None, wght=500)",
        new_x="LMARGIN",
        new_y="NEXT",
    )

    assert_pdf_equal(pdf, HERE / "variable_font_with_one_modified_axis.pdf", tmp_path)


def test_invalid_axes_dict():
    pdf = FPDF()
    pdf.add_page()

    # Value for "" is a string instead of a dict.
    invalid_variations_dict = {"B": ""}

    with pytest.raises(ValueError) as error:
        pdf.add_font(
            "Roboto Variable",
            "",
            HERE / "Roboto-Variable.ttf",
            variations=invalid_variations_dict,
        )
        assert error == "Variations dictionary is invalid"


def test_variations_not_dictionary():
    pdf = FPDF()
    pdf.add_page()

    # variations is not a dict
    invalid_variations_list = []
    with pytest.raises(TypeError) as error:
        pdf.add_font(
            "Roboto Variable",
            "",
            HERE / "Roboto-Variable.ttf",
            variations=invalid_variations_list,
        )
        assert error == "Variations, if specified, must be a dictionary"


def test_variations_are_axes(tmp_path):
    pdf = FPDF()
    pdf.add_page()

    # Provide axes dictionary instead of variations
    # and also provide a style.
    pdf.add_font(
        "Roboto Variable",
        "B",
        HERE / "Roboto-Variable.ttf",
        variations={"wght": 600},
    )

    pdf.set_font("Roboto Variable", "B", 20)

    pdf.cell(
        w=pdf.epw,
        text="Roboto Variable Bold(wdth=600)",
        new_x="LMARGIN",
        new_y="NEXT",
    )

    assert_pdf_equal(pdf, HERE / "variable_font_variations_are_axes.pdf", tmp_path)


def test_multicell_and_markdown(tmp_path):
    pdf = FPDF()
    pdf.add_page()

    # Add font with regular and bold style, but render
    # bold text using markdown and multi_cell() method.
    pdf.add_font(
        "Roboto Variable",
        "",
        HERE / "Roboto-Variable.ttf",
        variations={"": {"wdth": 100, "wght": 400}, "B": {"wght": 600}},
    )
    pdf.set_font("Roboto Variable", "", 20)

    pdf.multi_cell(
        w=pdf.epw,
        text="Regular and **Bold** text",
        markdown=True,
        new_x="LMARGIN",
        new_y="NEXT",
    )
    assert_pdf_equal(pdf, HERE / "variable_font_with_markdown.pdf", tmp_path)



from fpdf import FPDF, FPDFException, TextMode
from fpdf.line_break import (
    Fragment,
    MultiLineBreak,
    CurrentLine,
    TextLine,
    TotalPagesSubstitutionFragment,
)
from fpdf.enums import Align, CharVPos
from fpdf.graphics_state import GraphicsState

import pytest


class FxFragment(Fragment):
    """Test fixture returning a predetermined width for each character.
    Argument "wdict" is a dict of "styles" of arbitrary names.
    The value of each style is a dict of characters and widths.
    """

    def __init__(self, characters, graphics_state, k, link, wdict=None):
        self.wdict = wdict
        super().__init__(characters, graphics_state, k, link)

    def get_character_width(self, character, print_sh=False, initial_cs=True):
        """Return the relevant width from "wdict"."""
        cw = self.wdict[self.font_style]
        return cw[character]


class FxTotalPagesSubstitutionFragment(FxFragment, TotalPagesSubstitutionFragment):
    pass


class FxFont:
    def __init__(self, cw):
        self.i = 1
        self.cw = cw

    def get_char_width(self, char):
        return self.cw[char] if char in self.cw else 2

    def get_text_width(self, text, *_):
        return (len(text), sum(self.get_char_width(c) for c in text))


def test_fragment_properties():
    """
    Make sure the accessor properties in Fragment() return the correct
    value as of the originating graphics state.
    """
    # pylint: disable=protected-access
    pdf = FPDF()
    font_family = "helvetica"
    font_style = "I"
    font_size = 22
    pdf.set_font(font_family, font_style, font_size)
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert frag.font == pdf.current_font, (
        f"frag.font ({frag.font['name']}/{frag.font['fontkey']})"
        f" != pdf.current_font ({pdf.current_font['name']}/{pdf.current_font['fontkey']})"
    )
    assert frag.is_ttf_font == pdf.is_ttf_font, (
        f"frag.is_ttf_font ({frag.is_ttf_font})"
        f" != pdf.unifontsubset ({pdf.is_ttf_font})"
    )
    assert frag.font_family == pdf.font_family, (
        f"frag.font_family ({frag.font_family})"
        f" != pdf.font_family ({pdf.font_family})"
    )
    assert (
        frag.font_style == pdf.font_style
    ), f"frag.font_style ({frag.font_style}) != pdf.font_style ({pdf.font_style})"
    assert (
        frag.font_size == pdf.font_size
    ), f"frag.font_size ({frag.font_size}) != pdf.font_size ({pdf.font_size})"
    assert (
        frag.underline == pdf.underline
    ), f"frag.underline ({frag.underline}) != pdf.underline ({pdf.underline})"
    pdf.set_font_size(44)
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert (
        frag.font_size == pdf.font_size
    ), f"frag.font_size ({frag.font_size}) != pdf.font_size ({pdf.font_size})"
    pdf.set_stretching(120)
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert frag.font_stretching == pdf.font_stretching, (
        f"frag.font_stretching ({frag.font_stretching})"
        f" != pdf.font_stretching ({pdf.font_stretching})"
    )
    pdf.set_char_spacing(4)
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert frag.char_spacing == pdf.char_spacing, (
        f"frag.char_spacing ({frag.char_spacing})"
        f" != pdf.char_spacing ({pdf.char_spacing})"
    )
    pdf.text_mode = TextMode.STROKE
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert (
        frag.text_mode == pdf.text_mode
    ), f"frag.text_mode ({frag.text_mode}) != pdf.text_mode ({pdf.text_mode})"
    pdf.underline = True
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert (
        frag.underline == pdf.underline
    ), f"frag.underline ({frag.underline}) != pdf.underline ({pdf.underline})"
    pdf.set_draw_color(0.1, 0.2, 0.3)
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert (
        frag.draw_color == pdf.draw_color
    ), f"frag.draw_color ({frag.draw_color}) != pdf.draw_color ({pdf.draw_color})"
    pdf.set_fill_color(0.3, 0.2, 0.1)
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert (
        frag.fill_color == pdf.fill_color
    ), f"frag.fill_color ({frag.fill_color}) != pdf.fill_color ({pdf.fill_color})"
    pdf.set_text_color(0.5)
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert (
        frag.text_color == pdf.text_color
    ), f"frag.text_color ({frag.text_color}) != pdf.text_color ({pdf.text_color})"
    pdf.set_line_width(0.5)
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert (
        frag.line_width == pdf.line_width
    ), f"frag.line_width ({frag.line_width}) != pdf.line_width ({pdf.line_width})"
    pdf.char_vpos = "SUP"
    frag = Fragment("example", pdf._get_current_graphics_state(), pdf.k)
    assert (
        frag.char_vpos == pdf.char_vpos
    ), f"frag.char_vpos ({frag.char_vpos}) != pdf.char_vpos ({pdf.char_vpos})"


def test_no_fragments():
    """
    There is no text provided to break into multiple lines
    expected behavior ->
        - call to `get_line_of_given_width` always returns None
    """
    char_width = 6
    test_width = char_width * 200
    multi_line_break = MultiLineBreak([], test_width, [0, 0])
    assert multi_line_break.get_line() is None
    multi_line_break = MultiLineBreak([], 100, [0, 0])
    assert multi_line_break.get_line() is None


_gs_normal = GraphicsState(
    font_style="normal",
    font_size_pt=12,
    font_family="helvetica",
    font_stretching=100,
    char_spacing=0,
    current_font=None,
    char_vpos=CharVPos.LINE,
    text_shaping=None,
)
_gs_bold = GraphicsState(
    font_style="bold",
    font_size_pt=12,
    font_family="helvetica",
    font_stretching=100,
    char_spacing=0,
    current_font=None,
    char_vpos=CharVPos.LINE,
    text_shaping=None,
)


def gs_with_font(graphics_state, cw):
    gs = graphics_state
    gs.current_font = FxFont(cw)
    return gs


def test_width_calculation():
    """
    Every character has different width
    """
    text = "abcd"
    char_width = 2
    alphabet = {
        "normal": {},
    }

    def _get_width(height):  # pylint: disable=unused-argument
        return max_width

    for i, char in enumerate(text):
        alphabet["normal"][char] = char_width + i
    alphabet["normal"][" "] = char_width
    gs = _gs_normal
    gs.current_font = FxFont(alphabet["normal"])
    fragments = [FxFragment(text, gs, 1, None, alphabet)]
    multi_line_break = MultiLineBreak(fragments, _get_width, [0, 0])

    # zero width returns empty line
    max_width = 0
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[],
        text_width=0,
        number_of_spaces=0,
        align=Align.L,
        height=0,
        max_width=max_width,
        trailing_nl=False,
    )
    assert res == exp

    # the first character has width of char_width.
    # request of 1 unit line raises an exception
    with pytest.raises(FPDFException):
        max_width = 1
        res = multi_line_break.get_line()

    # get other characters one by one
    for i, char in enumerate(text):
        max_width = char_width + i
        res = multi_line_break.get_line()
        exp = TextLine(
            fragments=[Fragment(char, _gs_normal, 1)],
            text_width=char_width + i,
            number_of_spaces=0,
            align=Align.L,
            height=12,
            max_width=max_width,
            trailing_nl=False,
        )
        assert res == exp

    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_single_space_in_fragment():
    """
    there is only one space character in the input text.
    expected behavior ->
        - first call to `get_line` contains space.
        - second call to `get_line` is None because there is no
            text left.
    """
    text = " "
    char_width = 6
    test_width = char_width * 10
    alphabet = {
        "normal": {},
    }
    gs = _gs_normal
    gs.current_font = FxFont(alphabet["normal"])
    fragments = [FxFragment(text, gs, 1, None, alphabet)]
    for char in text:
        alphabet["normal"][char] = char_width
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=fragments,
        text_width=char_width,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_single_soft_hyphen_in_fragment():
    """
    there is only one soft hyphen character in the input text.
    expected behavior ->
        - call to `get_line` always returns None, because soft
          hyphen doesn't break a word
    """
    alphabet = {
        "normal": {"\u002d": 500},
    }
    text = "\u00ad"
    char_width = 6
    test_width = char_width * 200
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            text, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_single_hard_hyphen_in_fragment():
    """
    there is only one hard hyphen character in the input text.
    expected behavior ->
        - first call to `get_line` contains hard hyphen.
        - second call to `get_line` is None because there is no
    """
    alphabet = {
        "normal": {"\u002d": 500},
    }
    text = "\u002d"
    char_width = 6
    test_width = char_width * 4
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            text, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=fragments,
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_real_hyphen_acts_differently_from_soft_hyphen():
    words = ["a", "b", "c", "d"]
    char_width = 6
    test_width = char_width * 4
    alphabet = {
        "normal": {"\u002d": char_width},
    }
    words_separated_by_soft_hyphen = "\u00ad".join(words)
    words_separated_by_hard_hyphen = "\u002d".join(words)
    for char in words_separated_by_soft_hyphen:
        alphabet["normal"][char] = char_width
    soft_hyphen_line_break = MultiLineBreak(
        [
            FxFragment(
                words_separated_by_soft_hyphen,
                gs_with_font(_gs_normal, alphabet["normal"]),
                1,
                None,
                alphabet,
            )
        ],
        test_width,
        [0, 0],
    )
    hard_hyphen_line_break = MultiLineBreak(
        [
            FxFragment(
                words_separated_by_hard_hyphen,
                gs_with_font(_gs_normal, alphabet["normal"]),
                1,
                None,
                alphabet,
            )
        ],
        test_width,
        [0, 0],
    )
    hh_res = soft_hyphen_line_break.get_line()
    sh_res = hard_hyphen_line_break.get_line()
    assert hh_res != sh_res
    hh_res = soft_hyphen_line_break.get_line()
    sh_res = hard_hyphen_line_break.get_line()
    assert hh_res != sh_res


def test_trailing_soft_hyphen():
    """
    fit one word and trailing soft-hyphen into the line with extremely large width.
    expected behavior ->
        - first call to `get_line` contains the word.
          soft hyphen is not included in the line.
        - second call to `get_line` is None because there is no
            text left.
    """
    text = "hello\u00ad"
    char_width = 6
    test_width = char_width * 10
    test_width_B = char_width * 5
    alphabet = {
        "normal": {"\u002d": char_width},
    }
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            text, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("hello", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=test_width_B,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_trailing_whitespace():
    """
    fit one word and trailing whitespace into the line with extremely large width.
    expected behavior ->
        - first call to `get_line` contains the word and the space.
        - second call to `get_line` is None because there is no
            text left.
    """
    text = "hello "
    char_width = 6
    test_width = char_width * 10
    test_width_B = char_width * 6
    alphabet = {
        "normal": {},
    }
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            text, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=fragments,
        text_width=test_width_B,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_two_words_one_line():
    """
    fit two words into the line with extremely large width.
    expected behavior ->
        - first call to `get_line` contains all words.
        - second call to `get_line` is None because there is no
            text left.
    """
    text = "hello world"
    char_width = 6
    test_width = char_width * 200
    test_width_B = char_width * 11
    alphabet = {
        "normal": {},
    }
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            text, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=fragments,
        text_width=test_width_B,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_two_words_one_line_justify():
    """
    fit two words into the line with extremely large width.
    expected behavior ->
        - first call to `get_line` contains all words.
            this line is expected to be unjustified, because it is the last
            line.
        - second call to `get_line` is None because there is no
            text left.
    """
    text = "hello world"
    char_width = 6
    test_width = char_width * 200
    test_width_B = char_width * 11
    alphabet = {
        "normal": {},
    }
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            text, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0], align=Align.J)
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=fragments,
        text_width=test_width_B,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_two_words_two_lines_break_by_space():
    """
    fit two words into the line that can fit only one word.
    expected behavior:
        - first call to `get_line` contains the first word.
        - second call to `get_line` contains the second word.
        - third call to `get_line` is None because there is no
            text left.
    """
    text = "hello world"
    char_width = 6
    test_width = char_width * 5
    alphabet = {
        "normal": {},
    }
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            text, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("hello", _gs_normal, 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("world", _gs_normal, 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_two_words_two_lines_break_by_space_justify():
    """
    fit two words into the line that can fit only one word.
    expected behavior:
        - first call to `get_line` contains the first word.
            Line is expected to be unjustified, because there are no spaces in
            the line.
        - second call to `get_line` contains the second word.
            Line is expected to be unjustified, because it is the last line.
        - third call to `get_line` is None because there is no
            text left.
    """
    text = "hello world"
    char_width = 6
    test_width = char_width * 5
    alphabet = {
        "normal": {},
    }
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [FxFragment(text, _gs_normal, 1, None, alphabet)]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])

    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("hello", _gs_normal, 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("world", _gs_normal, 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_four_words_two_lines_break_by_space():
    """
    fit two words into the line that can fit only one word.
    expected behavior:
        - first call to `get_line` contains the first two words.
        - second call to `get_line` contains the second two words.
        - third call to `get_line` is None because there is no
            text left.
    """
    first_line_text = "hello world"
    second_line_text = "hello world"
    char_width = 6
    test_width_A = char_width * 12
    test_width_AA = char_width * 11
    text = " ".join([first_line_text, second_line_text])
    alphabet = {
        "normal": {},
    }
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [FxFragment(text, _gs_normal, 1, None, alphabet)]
    multi_line_break = MultiLineBreak(fragments, test_width_A, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment(first_line_text, _gs_normal, 1)],
        text_width=test_width_AA,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width_A,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment(second_line_text, _gs_normal, 1)],
        text_width=test_width_AA,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width_A,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_four_words_two_lines_break_by_space_justify():
    """
    fit two words into the line that can fit only one word.
    expected behavior:
        - first call to `get_line` contains the first two words.
            Line is expected to be justified.
        - second call to `get_line` contains the second two words.
            Line is expected to be unjustified, because it is the last line.
        - third call to `get_line` is None because there is no
            text left.
    """
    first_line_text = "hello world"
    second_line_text = "hello world"
    char_width = 6
    test_width_A = char_width * 12
    test_width_AA = char_width * 11
    text = " ".join((first_line_text, second_line_text))
    alphabet = {
        "normal": {},
    }
    for char in text:
        alphabet["normal"][char] = char_width
    fragments = [FxFragment(text, _gs_normal, 1, None, alphabet)]
    multi_line_break = MultiLineBreak(fragments, test_width_A, [0, 0], align=Align.J)
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment(first_line_text, _gs_normal, 1)],
        text_width=test_width_AA,
        number_of_spaces=1,
        align=Align.J,
        height=12,
        max_width=test_width_A,
        trailing_nl=False,
    )
    print(res)
    print(exp)
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment(second_line_text, _gs_normal, 1)],
        text_width=test_width_AA,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width_A,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_break_fragment_into_two_lines():
    """
    There are multiple fragments with different styles.
    This test breaks one fragment between two lines.
    """
    char_width = 6
    charB_width = 12
    test_width_A = char_width * 10
    test_width_B = char_width * 16
    test_width_BB = char_width * 15
    alphabet = {
        "normal": {},
        "bold": {},
    }
    first_line_text = "one "
    second_line_text = "two three"
    third_line_text = " four"
    text = "".join((first_line_text, second_line_text, third_line_text))
    for char in text:
        alphabet["normal"][char] = char_width
        alphabet["bold"][char] = charB_width

    fragments = [
        FxFragment(
            first_line_text,
            gs_with_font(_gs_normal, alphabet["normal"]),
            1,
            None,
            alphabet,
        ),
        FxFragment(
            second_line_text,
            gs_with_font(_gs_bold, alphabet["bold"]),
            1,
            None,
            alphabet,
        ),
        FxFragment(
            third_line_text,
            gs_with_font(_gs_normal, alphabet["normal"]),
            1,
            None,
            alphabet,
        ),
    ]

    def _get_width(height):  # pylint: disable=unused-argument
        return max_width

    multi_line_break = MultiLineBreak(fragments, _get_width, [0, 0])
    max_width = test_width_A
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment(first_line_text, _gs_normal, 1),
            Fragment("two", _gs_bold, 1),
        ],
        text_width=test_width_A,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width_A,
        trailing_nl=False,
    )
    assert res == exp
    max_width = test_width_B
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("three", _gs_bold, 1),
            Fragment(third_line_text, _gs_normal, 1),
        ],
        text_width=test_width_BB,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width_B,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_break_fragment_into_two_lines_justify():
    """
    There are multiple fragments with different styles.
    This test breaks one fragment between two lines.
    """
    char_width = 6
    charB_width = 12
    test_width_A = char_width * 10
    test_width_B = char_width * 16
    test_width_BB = char_width * 15
    alphabet = {
        "normal": {},
        "bold": {},
    }
    first_line_text = "one "
    second_line_text = "two three"
    third_line_text = " four"
    text = "".join((first_line_text, second_line_text, third_line_text))
    for char in text:
        alphabet["normal"][char] = char_width
        alphabet["bold"][char] = charB_width

    fragments = [
        FxFragment(
            first_line_text,
            gs_with_font(_gs_normal, alphabet["normal"]),
            1,
            None,
            alphabet,
        ),
        FxFragment(
            second_line_text,
            gs_with_font(_gs_bold, alphabet["bold"]),
            1,
            None,
            alphabet,
        ),
        FxFragment(
            third_line_text,
            gs_with_font(_gs_normal, alphabet["normal"]),
            1,
            None,
            alphabet,
        ),
    ]

    def _get_width(height):  # pylint: disable=unused-argument
        return max_width

    multi_line_break = MultiLineBreak(fragments, _get_width, [0, 0], align=Align.J)
    max_width = test_width_A
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment(first_line_text, _gs_normal, 1),
            Fragment("two", _gs_bold, 1),
        ],
        text_width=test_width_A,
        number_of_spaces=1,
        align=Align.J,
        height=12,
        max_width=test_width_A,
        trailing_nl=False,
    )
    assert res == exp
    max_width = test_width_B
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("three", _gs_bold, 1),
            Fragment(third_line_text, _gs_normal, 1),
        ],
        text_width=test_width_BB,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width_B,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_soft_hyphen_break():
    """
    all characters are separated by soft-hyphen
    expected behavior - there is a hard hyphen at the end of every line,
    except of the last one
    """
    char_width = 6
    test_width = char_width * 5
    test_width_A = char_width * 4.4
    test_width_AA = char_width * 4
    test_width_B = char_width * 2
    alphabet = {"normal": {"\u002d": char_width}}
    long_string = "\u00ad".join("abcdefghijklmnop")
    for char in long_string:
        alphabet["normal"][char] = char_width

    fragments = [
        FxFragment(
            long_string, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]

    def _get_width(height):  # pylint: disable=unused-argument
        return max_width

    multi_line_break = MultiLineBreak(fragments, _get_width, [0, 0])
    max_width = test_width
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("abcd\u002d", gs_with_font(_gs_normal, alphabet["normal"]), 1)
        ],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    print(res)
    print(exp)
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("efgh\u002d", gs_with_font(_gs_normal, alphabet["normal"]), 1)
        ],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    max_width = test_width_A
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("ijk\u002d", gs_with_font(_gs_normal, alphabet["normal"]), 1)
        ],
        text_width=test_width_AA,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width_A,
        trailing_nl=False,
    )
    assert res == exp
    max_width = test_width_B
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("l\u002d", gs_with_font(_gs_normal, alphabet["normal"]), 1)
        ],
        text_width=test_width_B,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width_B,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("m\u002d", gs_with_font(_gs_normal, alphabet["normal"]), 1)
        ],
        text_width=test_width_B,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width_B,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("n\u002d", gs_with_font(_gs_normal, alphabet["normal"]), 1)
        ],
        text_width=test_width_B,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width_B,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("op", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=test_width_B,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width_B,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_soft_hyphen_break_justify():
    """
    all characters are separated by soft-hyphen
    expected behavior - there is a hard hyphen at the end of every line,
    except of the last one
    """
    char_width = 6
    test_width = char_width * 6
    last_width = char_width * 5
    alphabet = {"normal": {"\u002d": char_width}}
    words = ["ab cd", "ef gh", "kl mn"]
    long_string = "\u00ad".join(words)
    for char in long_string:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            long_string, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0], align=Align.J)
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("ab cd\u002d", gs_with_font(_gs_normal, alphabet["normal"]), 1)
        ],
        text_width=test_width,
        number_of_spaces=1,
        align=Align.J,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[
            Fragment("ef gh\u002d", gs_with_font(_gs_normal, alphabet["normal"]), 1)
        ],
        text_width=test_width,
        number_of_spaces=1,
        align=Align.J,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("kl mn", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=last_width,
        number_of_spaces=1,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_explicit_break():
    """
    There is an explicit break character after every character
    Expected behavior:
        `get_line` returns single character on every call
    """
    char_width = 6
    test_width = char_width * 5
    alphabet = {
        "normal": {},
    }
    long_string = "\n".join("abcd")
    for char in long_string:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            long_string, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("a", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=True,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("b", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=True,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("c", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=True,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("d", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_explicit_break_justify():
    """
    There is an explicit break character after every character
    Expected behavior:
        `get_line` returns single character on every call,
        returned lines are expected to be unjustified
    """
    char_width = 6
    test_width = char_width * 5
    alphabet = {
        "normal": {},
    }
    long_string = "\n".join("abcd")
    for char in long_string:
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            long_string, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0], align=Align.J)
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("a", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=True,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("b", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=True,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("c", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=True,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("d", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_single_word_doesnt_fit_into_width():
    """
    There is a single word that doesn't fit into requested line
    Expected behavior:
        `get_line` as much characters as can fit into user
        provided width.
    """
    alphabet = {
        "normal": {},
    }
    long_string = "abcdefghijklmnop"
    char_width = 6
    test_width = char_width * 5
    for char in long_string:
        # glyph space units
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            long_string, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0])
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("abcde", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    print(res)
    print(exp)
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("fghij", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("klmno", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("p", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    assert res == exp
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_single_word_doesnt_fit_into_width_justify():
    """
    There is a single word that doesn't fit into requested line
    Expected behavior:
        `get_line` as much characters as can fit into user
        provided width. returned lines are expected to be unjustified
    """
    char_width = 6
    test_width = char_width * 5
    alphabet = {
        "normal": {},
    }
    long_string = "abcdefghijklmnop"
    for char in long_string:
        # glyph space units
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            long_string, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]
    multi_line_break = MultiLineBreak(fragments, test_width, [0, 0], align=Align.J)
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("abcde", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("fghij", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("klmno", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=test_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=[Fragment("p", gs_with_font(_gs_normal, alphabet["normal"]), 1)],
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,
        height=12,
        max_width=test_width,
        trailing_nl=False,
    )
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_last_line_no_justify():
    """
    Make sure that the last line is not justified.
    """
    alphabet = {
        "normal": {},
    }
    long_string = "a"
    char_width = 6
    for char in long_string:
        # glyph space units
        alphabet["normal"][char] = char_width
    fragments = [
        FxFragment(
            long_string, gs_with_font(_gs_normal, alphabet["normal"]), 1, None, alphabet
        )
    ]

    def _get_width(height):  # pylint: disable=unused-argument
        return max_width

    multi_line_break = MultiLineBreak(fragments, _get_width, [0, 0], align=Align.J)
    max_width = char_width * 5
    res = multi_line_break.get_line()
    exp = TextLine(
        fragments=fragments,
        text_width=char_width,
        number_of_spaces=0,
        align=Align.L,  # !
        height=12,
        max_width=max_width,
        trailing_nl=False,
    )
    assert res == exp
    max_width = char_width
    res = multi_line_break.get_line()
    exp = None
    assert res == exp


def test_trim_trailing_spaces():
    "Check special cases in CurrentLine method."
    # pylint: disable=protected-access,assignment-from-none
    pdf = FPDF()
    pdf.set_font("helvetica")
    cl = CurrentLine(pdf.w)
    # Result: None - if cl.fragments is empty to begin with.
    res = cl.trim_trailing_spaces()
    assert res is None
    # Result: None - if cl.fragments is empty after trimming trailing spaces.
    frag = Fragment(" ", pdf._get_current_graphics_state(), pdf.k)
    cl.fragments = [frag]
    res = cl.trim_trailing_spaces()
    assert res is None


def test_line_break_no_initial_newline():  # issue-847
    text = "X" * 50
    alphabet = {"normal": {}}
    alphabet["normal"]["X"] = 4.7
    fragments = [FxFragment(text, _gs_normal, 1, None, alphabet)]
    multi_line_break = MultiLineBreak(fragments, 188, [0, 0])
    text_line = multi_line_break.get_line()
    assert text_line.fragments


def test_substitution_fragments_are_not_merged_with_base_fragments():
    """
    There are three fragments, all of which fit on a single line. Normally, they would be merged into one fragment.
    However, a substitution fragment cannot be merged with a base fragment, otherwise,
    characters from the base fragment would be lost after replacement.

    Expected behavior:
        The fragments are not merged.
    """
    pdf = FPDF()
    assert pdf.str_alias_nb_pages
    alias = pdf.str_alias_nb_pages

    char_width = 6
    line_width = 200 * char_width
    text_before = "before"
    text_after = "after"

    alphabet = {
        "normal": {},
    }
    for char in text_before + alias + text_after:
        alphabet["normal"][char] = char_width

    graphics_state = gs_with_font(_gs_normal, alphabet["normal"])

    fragment_before = FxFragment(text_before, graphics_state, 1, None, alphabet)
    fragment_after = FxFragment(text_after, graphics_state, 1, None, alphabet)
    substitution_fragment = FxTotalPagesSubstitutionFragment(
        alias, graphics_state, 1, None, alphabet
    )

    multi_line_break = MultiLineBreak(
        [fragment_before, substitution_fragment, fragment_after],
        line_width,
        [0, 0],
    )

    line = multi_line_break.get_line()
    assert line.fragments == [fragment_before, substitution_fragment, fragment_after]

    assert multi_line_break.get_line() is None



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal

import pytest

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"


def test_multi_cell_markdown(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=32)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum dolor amet, **consectetur adipiscing** elit,"
        " sed do eiusmod __tempor incididunt__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown.pdf", tmp_path)


def test_multi_cell_markdown_strikethrough(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=32)
    pdf.multi_cell(w=pdf.epw, text="~~strikethrough~~", markdown=True)
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_strikethrough.pdf", tmp_path)


def test_multi_cell_markdown_escaped(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=32)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum \\ dolor amet, \\**consectetur adipiscing\\** elit,"
        " sed do eiusmod \\\\__tempor incididunt\\\\__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_escaped.pdf", tmp_path)


def test_multi_cell_markdown_with_ttf_fonts(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", "B", FONTS_DIR / "Roboto-Bold.ttf")
    pdf.add_font("Roboto", "I", FONTS_DIR / "Roboto-Italic.ttf")
    pdf.set_font("Roboto", size=32)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum dolor, **consectetur adipiscing** elit,"
        " eiusmod __tempor incididunt__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_with_ttf_fonts.pdf", tmp_path)


def test_multi_cell_markdown_with_ttf_fonts_escaped(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", "B", FONTS_DIR / "Roboto-Bold.ttf")
    pdf.add_font("Roboto", "I", FONTS_DIR / "Roboto-Italic.ttf")
    pdf.set_font("Roboto", size=32)
    text = (  # Some text where styling occur over line breaks:
        "Lorem ipsum \\ dolor, \\**consectetur adipiscing\\** elit,"
        " eiusmod \\\\__tempor incididunt\\\\__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(
        w=pdf.epw, text=text, markdown=True
    )  # This is tricky to get working well
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(
        pdf, HERE / "multi_cell_markdown_with_ttf_fonts_escaped.pdf", tmp_path
    )


def test_multi_cell_markdown_missing_ttf_font():
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.add_font(fname=FONTS_DIR / "Roboto-Regular.ttf")
    pdf.set_font("Roboto-Regular", size=60)
    with pytest.raises(fpdf.FPDFException) as error:
        pdf.multi_cell(w=pdf.epw, text="**Lorem Ipsum**", markdown=True)
    expected_msg = "Undefined font: roboto-regularB - Use built-in fonts or FPDF.add_font() beforehand"
    assert str(error.value) == expected_msg


def test_multi_cell_markdown_with_fill_color(tmp_path):  # issue 348
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=10)
    pdf.set_fill_color(255, 0, 0)
    pdf.multi_cell(
        50, markdown=True, text="aa bb cc **dd ee dd ee dd ee dd ee dd ee dd ee**"
    )
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_with_fill_color.pdf", tmp_path)


def test_multi_cell_markdown_justified(tmp_path):  # issue 327
    pdf = fpdf.FPDF()
    pdf.add_page()
    for font in ("Helvetica", "Courier"):
        pdf.set_font(family=font, size=12)
        pdf.set_y(pdf.y + 3)
        pdf.multi_cell(
            190,
            markdown=True,
            align="J",
            text=(
                "Lorem **ipsum** dolor sit amet, **consectetur** adipiscing elit, "
                "sed do eiusmod tempor incididunt ut labore et dolore magna "
                "aliqua. Ut enim ad minim veniam, __quis__ nostrud exercitation "
                "ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis "
                "aute irure dolor in reprehenderit in voluptate velit esse cillum "
                "dolore eu fugiat nulla pariatur. Excepteur sint occaecat "
                "cupidatat non proident, sunt in culpa qui officia deserunt "
                "mollit anim id est laborum."
            ),
        )
        pdf.set_x(10)
        pdf.set_y(pdf.y + 5)
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_justified.pdf", tmp_path)


def test_multi_cell_markdown_link(tmp_path):
    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()
    pdf.multi_cell(
        pdf.epw,
        text="**Start** [One Page Dungeon Context](https://www.dungeoncontest.com/) __End__",
        markdown=True,
    )
    assert_pdf_equal(pdf, HERE / "multi_cell_markdown_link.pdf", tmp_path)



import pytest

from fpdf.enums import TextEmphasis


def test_text_emphasis_coerce():
    assert TextEmphasis.coerce("B") == TextEmphasis.B
    assert TextEmphasis.coerce("BOLD") == TextEmphasis.B
    assert TextEmphasis.coerce("I") == TextEmphasis.I
    assert TextEmphasis.coerce("italics") == TextEmphasis.I
    assert TextEmphasis.coerce("U") == TextEmphasis.U
    assert TextEmphasis.coerce("Underline") == TextEmphasis.U
    with pytest.raises(ValueError):
        assert TextEmphasis.coerce("BXXX")
    assert (
        TextEmphasis.coerce("BIU") == TextEmphasis.B | TextEmphasis.I | TextEmphasis.U
    )


def test_text_emphasis_style():
    assert TextEmphasis.coerce("B").style == "B"
    assert TextEmphasis.coerce("IB").style == "BI"
    assert TextEmphasis.coerce("BIU").style == "BIU"


def test_text_emphasis_add():
    assert TextEmphasis.B.add(TextEmphasis.I).add(
        TextEmphasis.U
    ) == TextEmphasis.coerce("BIU")


def test_text_emphasis_remove():
    assert (
        TextEmphasis.coerce("BIU").remove(TextEmphasis.B).remove(TextEmphasis.I)
        == TextEmphasis.U
    )



#!/usr/bin/env python3
# Script Dependencies:
#    openxlsx
# USAGE: ./xlsx2table.py color_srgb.xlsx
import sys
from fpdf import FPDF, FontFace
from fpdf.drawing_primitives import color_from_hex_string
from openpyxl import load_workbook

pdf = FPDF()
pdf.add_page()
pdf.set_font("Times", size=22)
wb = load_workbook(sys.argv[1])
ws = wb.active
with pdf.table() as table:
    for i, row in enumerate(ws.rows):
        style = None
        if i > 0:
            # We color the row based on the hexadecimal code in the 2nd column:
            style = FontFace(fill_color=color_from_hex_string(row[1]))
        table.row([cell.value for cell in row], style=style)
pdf.output("from-xlsx.pdf")
